'''
Created on 2015/8/18

:author: hubo
'''

from vlcp.config import defaultconfig
from vlcp.protocol import Protocol
from vlcp.event import Event, withIndices, Stream, ConnectionControlEvent, ConnectionWriteEvent, CBQueue, PollEvent
from vlcp.event.core import syscall_clearremovequeue, PollEvent
from collections import deque
import logging
import re
import time
from vlcp.event.stream import MemoryStream, StreamDataEvent
import zlib

@withIndices('state', 'connection', 'connmark', 'createby')
class HttpConnectionStateEvent(Event):
    """
    HTTP connection state changed
    """
    # Connection established from the client side
    CLIENT_CONNECTED = 'clientconnect'
    # Connection down from the client side
    CLIENT_CLOSE = 'clientclose'
    # Cannot connect to server
    CLIENT_NOTCONNECTED = 'clientnotconnected'
    # A connection is established to server side
    SERVER_CONNECTED = 'serverconnect'
    # A connection is down from the server side
    SERVER_CLOSE = 'serverclose'
    

@withIndices('host', 'path', 'method', 'connection', 'connmark', 'xid', 'createby')
class HttpRequestEvent(Event):
    """
    A HTTP request is received from the connection
    """
    canignore = False
    def canignorenow(self):
        return not self.connection.connected or self.connection.connmark != self.connmark

@withIndices('connection', 'connmark', 'xid', 'isfinal', 'iserror')
class HttpResponseEvent(Event):
    """
    A HTTP response is received from the connection
    """
    pass

@withIndices('connection', 'connmark', 'xid', 'keepalive')
class HttpResponseEndEvent(Event):
    """
    A HTTP response is fully received
    """
    pass

@withIndices('connection', 'connmark', 'type')
class HttpStateChange(Event):
    NEXTINPUT = 'nextinput'
    NEXTOUTPUT = 'nextoutput'

@withIndices('stream')
class HttpTrailersReceived(Event):
    """
    Trailers are received on an HTTPResponseStream
    """
    pass

try:
    _long = long
except:
    _long = int

class HttpProtocolException(Exception):
    """
    Critical protocol break on HTTP connections
    """
    pass

class HttpConnectionClosedException(HttpProtocolException):
    """
    Connection is closed
    """
    pass

linedelimiter = re.compile(br'\r\n')
statuscheck = re.compile(br'HTTP/\d\.\d[ \t]')
statuscheckshort = re.compile(br'H(?:T(?:T(?:P(?:/(?:\d(?:\.(?:\d[ \t]?)?)?)?)?)?)?)?$')
token = br"[\!#$%&\'*+\-\.^_`|~0-9a-zA-Z]+"
token_r = re.compile(br"^[\!#$%&\'*+\-\.^_`|~0-9a-zA-Z]+$")
requestline = re.compile(br'^(' + token + br')[ \t]+([^ \t]+)[ \t]+HTTP/(\d\.\d)$')
statusline = re.compile(br'^HTTP/(\d\.\d)[ \t]+(\d{3})[ \t]+(.*)$')
headerline = re.compile(br'^(' + token + br')\:[ \t]*(.*?)[ \t]*$')
chunkedheaderline = re.compile(br'^([0-9a-zA-Z]+)(?:;.*)?$')
standard_status = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',
    208: 'Already Reported',
    226: 'IM Used',
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    306: '(Unused)',
    307: 'Temporary Redirect',
    308: 'Permanent Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Payload Too Large',
    414: 'URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Range Not Satisfiable',
    417: 'Expectation Failed',
    421: 'Misdirected Request',
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    426: 'Upgrade Required',
    428: 'Precondition Required',
    429: 'Too Many Requests',
    431: 'Request Header Fields Too Large',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    506: 'Variant Also Negotiates',
    507: 'Insufficient Storage',
    508: 'Loop Detected',
    510: 'Not Extended',
    511: 'Network Authentication Required'
}
weekdayname = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
monthname = [None,
             'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
             'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
# From BaseHTTPServer library
def date_time_string(timestamp=None):
    """Return the current date and time formatted for a message header."""
    if timestamp is None:
        timestamp = time.time()
    year, month, day, hh, mm, ss, wd, y, z = time.gmtime(timestamp)
    s = "%s, %02d %3s %4d %02d:%02d:%02d GMT" % (
            weekdayname[wd],
            day, monthname[month], year,
            hh, mm, ss)
    return s

# From cgi library
def escape_b(s, quote=True):
    '''Replace special characters "&", "<" and ">" to HTML-safe sequences.
    If the optional flag quote is true, the quotation mark character (")
    is also translated.'''
    s = s.replace(b"&", b"&amp;") # Must be done first!
    s = s.replace(b"<", b"&lt;")
    s = s.replace(b">", b"&gt;")
    if quote:
        s = s.replace(b'"', b"&quot;")
    return s

def escape(s, quote=True):
    '''Replace special characters "&", "<" and ">" to HTML-safe sequences.
    If the optional flag quote is true, the quotation mark character (")
    is also translated.'''
    s = s.replace("&", "&amp;") # Must be done first!
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    if quote:
        s = s.replace('"', "&quot;")
    return s

def normalizeHeader(headername):
    return headername.lower()

specialdisplayheaders = {b'etag':b'ETag', }

def displayHeader(headername):
    return b'-'.join(p.capitalize() for p in headername.split(b'-'))

def _createstatus(status):
    if isinstance(status, int) or isinstance(status, _long):
        status = str(status) + ' ' + standard_status.get(status, 'User-defined')
    if not isinstance(status, bytes):
        status = status.encode('ascii')
    return status


@defaultconfig
class Http(Protocol):
    '''
    Basic HTTP/1.1 protocol
    Base on RFC723x, which are more strict than RFC2616
    '''
    _default_persist = False
    # Limit the pipelining requests that are executed in parallel.
    # Disable pipelining by setting pipelinelimit = 1
    _default_pipelinelimit = 10
    # Maximum length for the first line of request or response
    _default_headlinelimit = 8192
    # Maximum length for a single header
    _default_headerlimit = 1048576
    # Maximum header count for a single request/response
    _default_headercountlimit = 1024
    # Maximum length for a chunked header; it is a special header type which is rarely used.
    _default_chunkedheaderlimit = 8192
    # Maximum chunked header count
    _default_chunkednumberlimit = 16
    # Enable HTTP/0.9, in which a response does not have headers
    _default_allownoheader = True
    # Default HTTP protocol
    _default_defaultversion = '1.1'
    # Drop the connection rather than sending a 400 Bad Request
    # when there is a protocol break in the request or response
    _default_fastfail = False
    # Enable TE: deflate header. Not all implements support this.
    _default_allowtransfercompress = False
    # Close a HTTP/1.1 keep-alive connection when idles for a specified time
    _default_idletimeout = 20
    # Close a connection when "Connection: close" is received after the specified time
    _default_closetimeout = 5
    # Print debugging logs
    _default_debugging = False
    # Use Expect: 100-continue for POST requests; not all server supports this
    _default_useexpect = False
    # If useexpect = True, Use Expect: 100-continue for POST requests with data larger than this size
    _default_useexpectsize = 4096
    # When 100 Continue response is not received for the specified time, send the data
    _default_expecttimeout = 2
    # default HTTP port
    _default_defaultport = 80
    # default HTTPS port
    _default_ssldefaultport = 443
    # Show error details in error response
    _default_showerrorinfo = False
    # Use unquoteplus instead of unquote, which means + is recognized as space ' '
    _default_unquoteplus = True
    _logger = logging.getLogger(__name__ + '.HTTP')
    # A dictionary to rewrite errors in HttpHandler to specified path (without 3xx redirect)
    _default_errorrewrite = {}
    # A dictionary to redirect errors in HttpHandler to specified path (with 302 Found)
    _default_errorredirect = {}
    # Priority for response end events
    _default_responseendpriority = 690
    # Default response headers that are sent from HTTP server to client
    _default_defaultresponseheaders = [(b'Content-Type', b'text/html'), (b'Server', b'VLCP HTTP Server'),
                                    (b'Vary', b'Accept-Encoding')]
    # Default request headers that are sent from HTTP client to server
    _default_defaultrequestheaders = [(b'Accept', b'*/*'), (b'User-Agent', b'VLCP HTTP Client')]
    _default_tcp_nodelay = True
    def __init__(self, server = True, defaultversion = None):
        '''
        Constructor
        '''
        Protocol.__init__(self)
        self.server = server
        self.connectioncount = 0
        self.connectionpersource = {}
        if defaultversion is not None:
            self.defaultversion = defaultversion
    def createmessagequeue(self, scheduler):
        if 'httprequest' not in scheduler.queue:
            #scheduler.queue.addSubQueue(self.messagepriority, HttpRequestEvent.createMatcher(), 'httprequest', None, None, CBQueue.AutoClassQueue.initHelper('connection', 100, subqueuelimit = self.pipelinelimit * 2 if self.pipelinelimit is not None else None))
            scheduler.queue.addSubQueue(self.messagepriority, HttpRequestEvent.createMatcher(), 'httprequest', None, None)
            scheduler.queue.addSubQueue(self.messagepriority, HttpConnectionStateEvent.createMatcher(), 'httpconnstate', None, None)
            #scheduler.queue.addSubQueue(self.messagepriority, HttpResponseEvent.createMatcher(), 'httpresponse', None, None, CBQueue.AutoClassQueue.initHelper('connection', 100, subqueuelimit = self.pipelinelimit * 2 if self.pipelinelimit is not None else None))
            scheduler.queue.addSubQueue(self.messagepriority, HttpResponseEvent.createMatcher(), 'httpresponse', None, None)
            scheduler.queue.addSubQueue(self.responseendpriority, HttpResponseEndEvent.createMatcher(), 'httpresponseend', None, None)
            #scheduler.queue.addSubQueue(self.messagepriority, HttpTrailersReceived.createMatcher(), 'httptrailer', None, None, CBQueue.AutoClassQueue.initHelper('stream', 100, subqueuelimit = None))
            scheduler.queue.addSubQueue(self.messagepriority, HttpTrailersReceived.createMatcher(), 'httptrailer', None, None)
            scheduler.queue.addSubQueue(self.writepriority, HttpStateChange.createMatcher(), 'httpstatechange')
    def init(self, connection):
        for m in Protocol.init(self, connection):
            yield m
        self.createmessagequeue(connection.scheduler)
        for m in self.reconnect_init(connection):
            yield m
    def _request_sender(self, connection):
        try:
            connmark = connection.connmark
            nextOutput = HttpStateChange.createMatcher(connection, 
                                                      connmark, 
                                                      HttpStateChange.NEXTOUTPUT)
            def requesterror(msg, method, path):
                self._logger.error('Illegal HTTP request: %s, connection = %r, request = %r', msg, connection, method + b' ' + path + b' HTTP/' + connection.http_localversion.encode('ascii'))
                raise HttpProtocolException(msg)
            writeclose = False
            while not writeclose:
                if not connection.http_requestbuffer:
                    yield (nextOutput,)
                else:
                    # (host, path, method, headers, stream)
                    host, path, method, headers, output, keepalive = connection.http_requestbuffer.popleft()
                    reqteinfo = (method, True, True, True)
                    connection.http_reqteinfo.append(reqteinfo)
                    xid = connection.xid
                    connection.xid += 1
                    # find out how we should use the output stream
                    transfer_chunked = False
                    transfer = None
                    try:
                        if method == b'CONNECT':
                            content_length = None
                        elif output is None:
                            content_length = 0
                        else:
                            cl = [(k,v) for k,v in headers if normalizeHeader(k) == b'content-length']
                            if len(cl) > 1:
                                requesterror('multiple content-length headers', method, path)
                            elif cl:
                                try:
                                    content_length = int(cl[0][1])
                                except:
                                    requesterror('invalid content-length header', method, path)
                            else:
                                if connection.http_localversion < '1.1':
                                    requesterror('Must have content-length for HTTP/1.0 request with message body')
                                else:
                                    transfer_chunked = True
                                    transfer = b'chunked'
                        # Output headline and headers
                        buffer = []
                        def write():
                            for m in connection.write(ConnectionWriteEvent(connection,
                                                                           connmark,
                                                                           data = b''.join(buffer))):
                                yield m
                            del buffer[:]
                        buffer.append(method + b' ' + path + b' HTTP/' +
                                      connection.http_localversion.encode('ascii') + b'\r\n')
                        if transfer is not None:
                            buffer.append(b'Transfer-Encoding: ' + transfer + b'\r\n')
                        for k,v in headers:
                            buffer.append(k + b': ' + v + b'\r\n')
                        expect = self.useexpect and output is not None and \
                            (transfer is not None or content_length is not None and content_length > self.useexpectsize)
                        if expect:
                            buffer.append(b'Expect: 100-continue\r\n')
                        buffer.append(b'\r\n')
                        for m in write():
                            yield m
                        if expect:
                            def waitForContinue():
                                expect100 = self.responsematcher(connection, xid)
                                while True:
                                    yield (expect100,)
                                    if connection.event.isfinal:
                                        break
                                    elif connection.event.statuscode == 100:
                                        break
                            for m in connection.executeWithTimeout(self.expecttimeout, waitForContinue()):
                                yield m
                            if not connection.timeout:
                                # There is a response
                                if connection.event.isfinal:
                                    # We have already received a response, do not send the body
                                    for m in connection.write(ConnectionWriteEvent(connection,
                                                                                       connmark,
                                                                                       data = b'',
                                                                                       EOF = True)):
                                        yield m
                                    writeclose = True
                                    break
                        # Send data
                        if output is not None:
                            while True:
                                try:
                                    for m in output.prepareRead(connection):
                                        yield m
                                    data = output.readonce()
                                except EOFError:
                                    if transfer_chunked:
                                        # Send chunked end
                                        buffer.append(b'0\r\n')
                                        # Trailers
                                        if hasattr(output, 'trailers'):
                                            for k,v in output.trailers:
                                                buffer.append(k + b': ' + v + b'\r\n')
                                        buffer.append(b'\r\n')
                                        if buffer:
                                            for m in write():
                                                yield m
                                        if not keepalive:
                                            for m in connection.write(ConnectionWriteEvent(connection,
                                                                                           connmark,
                                                                                           data = b'',
                                                                                           EOF = True)):
                                                yield m
                                            writeclose = True
                                            break 
                                    elif content_length is None:
                                        for m in connection.write(ConnectionWriteEvent(connection,
                                                                                           connmark,
                                                                                           data = b'',
                                                                                           EOF = True)):
                                            yield m
                                        writeclose = True
                                        break
                                    elif content_length > 0:
                                        raise
                                    break
                                else:
                                    if transfer_chunked:
                                        if data:
                                            buffer.append(hex(len(data))[2:].encode('ascii') + b'\r\n')
                                            buffer.append(data)
                                            buffer.append(b'\r\n')
                                            for m in write():
                                                yield m
                                    else:
                                        if content_length is not None:
                                            if len(data) > content_length:
                                                data = data[0:content_length]
                                                
                                        for m in connection.write(ConnectionWriteEvent(connection,
                                                                                       connmark,
                                                                                       data = data)):
                                            yield m
                                        if content_length is not None:
                                            content_length -= len(data)
                                            if content_length <= 0:
                                                break
                    finally:
                        if output is not None:
                            output.close(connection.scheduler)
        except:
            for m in connection.reset(True, connmark):
                yield m
            raise            
    def _response_sender(self, connection):
        try:
            connmark = connection.connmark
            nextInput = HttpStateChange.createMatcher(connection, 
                                                      connmark, 
                                                      HttpStateChange.NEXTINPUT)
            nextOutput = HttpStateChange.createMatcher(connection, 
                                                      connmark, 
                                                      HttpStateChange.NEXTOUTPUT)
            writeclose = False
            while True:
                if writeclose:
                    if connection.http_remoteversion >= '1.1':
                        for m in connection.waitWithTimeout(self.closetimeout):
                            yield m
                    for m in connection.shutdown(False, connmark):
                        yield m
                    break
                elif connection.xid == connection.http_responsexid and connection.http_idle:
                    if connection.http_keepalive:
                        # Wait for new requests
                        for m in connection.waitWithTimeout(self.idletimeout, nextInput):
                            yield m
                    else:
                        if connection.http_remoteversion >= '1.1':
                            for m in connection.waitWithTimeout(self.closetimeout):
                                yield m
                        else:
                            connection.timeout = True
                    if connection.timeout:
                        for m in connection.shutdown(False, connmark):
                            yield m
                        break
                elif not connection.http_responsexid in connection.http_responsebuffer or\
                        not connection.http_responsebuffer[connection.http_responsexid]:
                    yield (nextOutput,)
                else:
                    resp = connection.http_responsebuffer[connection.http_responsexid].pop(0)
                    # find out how we should use the output stream
                    transfer_deflate = False
                    transfer_chunked = False
                    transfer = None
                    try:
                        output = None
                        if resp.status[:1] == b'1':
                            # 1xx does not have body
                            output = None
                            transfer = None
                            if resp.outputstream is not None:
                                resp.outputstream.close(connection.scheduler, True)
                        else:
                            teinfo = connection.http_reqteinfo.popleft()
                            if teinfo[0] == b'HEAD':
                                # HEAD response has no body
                                output = None
                                transfer = None
                                if resp.outputstream is not None:
                                    resp.outputstream.close(connection.scheduler, True)
                            elif teinfo[0] == b'CONNECT':
                                output = resp.outputstream
                                transfer = None
                                content_length = None
                            elif resp.status[:4] in (b'204 ', b'304 '):
                                # 204 and 304 response has no body
                                output = None
                                transfer = None
                                if resp.outputstream is not None:
                                    resp.outputstream.close(connection.scheduler, True)
                            else:
                                output = resp.outputstream
                                cl = [(k,v) for k,v in resp.headers if normalizeHeader(k) == b'content-length']
                                if len(cl) > 1:
                                    resp = self.createErrorResponse(connection, connection.http_responsexid, 500)
                                    cl = [(k,v) for k,v in resp.headers if normalizeHeader(k) == b'content-length']
                                    content_length = int(cl[0][1])
                                elif cl:
                                    try:
                                        content_length = int(cl[0][1])
                                    except Exception:
                                        resp = self.createErrorResponse(connection, connection.http_responsexid, 500)
                                        cl = [(k,v) for k,v in resp.headers if normalizeHeader(k) == b'content-length']
                                        content_length = int(cl[0][1])
                                elif output is None:
                                    content_length = 0
                                    resp.headers.append((b'Content-Length', b'0'))
                                else:
                                    content_length = None
                                if content_length is None and teinfo[3]:
                                    if teinfo[2] and self.allowtransfercompress and not getattr(resp, 'disabledeflate', False):
                                        transfer = b'deflate, chunked'
                                        transfer_deflate = True
                                        deflateobj = zlib.compressobj()
                                    else:
                                        transfer = b'chunked'
                                        transfer_deflate = False
                                    transfer_chunked = True
                                else:
                                    transfer = None
                                    transfer_chunked = False
                                    transfer_deflate = False
                            connection.http_responsexid += 1
                            # More requests
                            while connection.http_requestbuffer and connection.xid - len(connection.http_requestbuffer) < connection.http_responsexid + self.pipelinelimit:
                                r = connection.http_requestbuffer.popleft()
                                for m in connection.waitForSend(r):
                                    yield m
                                
                        # Output headline and headers
                        buffer = []
                        def write():
                            for m in connection.write(ConnectionWriteEvent(connection,
                                                                           connmark,
                                                                           data = b''.join(buffer))):
                                yield m
                            del buffer[:]
                        buffer.append(b'HTTP/' + connection.http_localversion.encode('ascii') + b' '
                                      + resp.status + b'\r\n')
                        if transfer is not None:
                            buffer.append(b'Transfer-Encoding: ' + transfer + b'\r\n')
                        for k,v in resp.headers:
                            buffer.append(k + b': ' + v + b'\r\n')
                        buffer.append(b'\r\n')
                        for m in write():
                            yield m
                        # Send data
                        if output is not None:
                            while True:
                                try:
                                    for m in output.prepareRead(connection):
                                        yield m
                                    data = output.readonce()
                                except EOFError:
                                    if transfer_chunked:
                                        if transfer_deflate:
                                            data = deflateobj.flush()
                                            if data:
                                                buffer.append(hex(len(data))[2:].encode('ascii'))
                                                buffer.append(b'\r\n')
                                                buffer.append(data)
                                                buffer.append(b'\r\n')
                                        # Send chunked end
                                        buffer.append(b'0\r\n')
                                        # Send trailers if any
                                        if resp.trailers is not None:
                                            for k,v in resp.trailers:
                                                buffer.append(k + b': ' + v + b'\r\n')
                                        buffer.append(b'\r\n')
                                        for m in write():
                                            yield m
                                    elif content_length is None or content_length > 0:
                                        for m in connection.write(ConnectionWriteEvent(connection,
                                                                                           connmark,
                                                                                           data = b'',
                                                                                           EOF = True)):
                                            yield m
                                        writeclose = True
                                        #if content_length is not None:
                                        #    raise
                                    break
                                else:
                                    if transfer_chunked:
                                        if transfer_deflate:
                                            data = deflateobj.compress(data)
                                        if data:
                                            buffer.append(hex(len(data))[2:].encode('ascii') + b'\r\n')
                                            buffer.append(data)
                                            buffer.append(b'\r\n')
                                            for m in write():
                                                yield m
                                    else:
                                        if content_length is not None:
                                            if len(data) > content_length:
                                                data = data[0:content_length]
                                        for m in connection.write(ConnectionWriteEvent(connection,
                                                                                       connmark,
                                                                                       data = data)):
                                            yield m
                                        if content_length is not None:
                                            content_length -= len(data)
                                            if content_length <= 0:
                                                break
                    finally:
                        if output is not None:
                            output.close(connection.scheduler)
        except:
            for m in connection.reset(True, connmark):
                yield m
            raise
    def reconnect_init(self, connection):
        connection.xid = 0
        connection.http_responsexid = 0
        connection.http_parsestage = 'headline'
        connection.http_reqteinfo = deque()
        connection.http_requestbuffer = deque()
        connection.http_responsebuffer = {}
        connection.http_localversion = self.defaultversion
        connection.http_remoteversion = None
        connection.http_currentstream = None
        connection.http_idle = True
        if connection.http_localversion >= '1.1':
            connection.http_keepalive = True
        else:
            connection.http_keepalive = False
        if self.server:
            connection.subroutine(self._response_sender(connection), name = 'http_sender')
        else:
            connection.subroutine(self._request_sender(connection), name = 'http_sender')
        state = HttpConnectionStateEvent(HttpConnectionStateEvent.SERVER_CONNECTED if self.server
                                         else HttpConnectionStateEvent.CLIENT_CONNECTED,
                                         connection, connection.connmark, self)
        for m in connection.waitForSend(state):
            yield m
    def _clearresponse(self, connection):
        if connection.http_currentstream is not None:
            for m in connection.http_currentstream.error(connection, ignoreexception = True):
                yield m
            connection.http_currentstream = None
        for i in range(connection.http_responsexid, connection.xid + 1):
            if i in connection.http_responsebuffer:
                for r in connection.http_responsebuffer[i]:
                    if r.outputstream is not None:
                        r.outputstream.close(connection.scheduler)
        connection.http_responsebuffer.clear()
    def closed(self, connection):
        connection.http_sender.close()
        connection.http_sender = None
        for m in Protocol.closed(self, connection):
            yield m
        connection.scheduler.ignore(HttpRequestEvent.createMatcher(None, None, None, connection))
        for m in connection.waitForSend(HttpConnectionStateEvent(HttpConnectionStateEvent.SERVER_CLOSE if self.server
                                                                 else HttpConnectionStateEvent.CLIENT_CLOSE, connection, connection.connmark, self)):
            yield m
        for m in self._clearresponse(connection):
            yield m
        connection.http_reqteinfo = None
        connection.http_requestbuffer = None
        connection.http_parsestage = 'end'
    def error(self, connection):
        connection.http_sender.close()
        for m in Protocol.error(self, connection):
            yield m
        connection.scheduler.ignore(HttpRequestEvent.createMatcher(None, None, None, connection))
        if self.debugging:
            self._logger.debug('Http connection is reset on %r', connection)
        for m in connection.waitForSend(HttpConnectionStateEvent(HttpConnectionStateEvent.SERVER_CLOSE if self.server
                                                                 else HttpConnectionStateEvent.CLIENT_CLOSE, connection, connection.connmark, self)):
            yield m
        for m in self._clearresponse(connection):
            yield m
        connection.http_reqteinfo = None
        connection.http_requestbuffer = None
    def _createResponseheaders(self, connection, xid, headers, status):
        defaultheaders = self.defaultresponseheaders
        if not connection.http_keepalive and xid == connection.xid - 1 and status[0] != b'1'[0]:
            newheaders = [(b'Connection', b'Close')]
        else:
            newheaders = [(b'Connection', b'Keep-Alive')]
        newheaders.append((b'Date', date_time_string().encode('ascii')))
        existingHeaders = set(normalizeHeader(k) for k,_ in headers)
        existingHeaders.add(b'date')
        existingHeaders.add(b'connection')
        for k,v in defaultheaders:
            if normalizeHeader(k) not in existingHeaders:
                newheaders.append((k, v))
        newheaders.extend((k,v) for k,v in headers if normalizeHeader(k) not in (b'connection', \
                          b'date', b'transfer-encoding'))
        return newheaders
    class _HttpResponse(object):
        def __init__(self, status, headers, outputstream, trailers = None):
            self.status = status
            self.headers = headers
            self.outputstream = outputstream
            self.trailers = trailers
    def createResponse(self, connection, xid, status, headers, outputstream):
        status = _createstatus(status)
        headers = self._createResponseheaders(connection, xid, headers, status)
        return self._HttpResponse(status, headers, outputstream)
    def createErrorResponse(self, connection, xid, status):
        status = _createstatus(status)
        content = b'<h1>' + escape_b(status) + b'</h1>'
        content_length = len(content)
        return self.createResponse(connection, xid, status, [(b'Content-Length', str(content_length).encode('ascii'))], MemoryStream(content))
    def createContinueResponse(self, connection, xid):
        return self.createResponse(connection, xid, b'100 Continue', [], None)
    class _HttpContinueStream(Stream):
        '''
        Send a 100-continue before reading data
        '''
        def __init__(self, protocol, connection, xid):
            Stream.__init__(self)
            self.http_continue = False
            self.http_protocol = protocol
            self.http_connection = connection
            self.http_xid = xid
        def prepareRead(self, container):
            if not self.http_continue:
                self.http_continue = True
                event = self.http_protocol.responseTo(self.http_connection, self.http_xid, 
                                              self.http_protocol.createContinueResponse(self.http_connection, self.http_xid))
                if event is not None:
                    for m in container.waitForSend(event):
                        yield m
            for m in Stream.prepareRead(self, container):
                yield m
    def responseTo(self, connection, xid, response):
        '''
        Return an event if notify is necessary
        '''
        if xid < connection.http_responsexid:
            raise HttpProtocolException('xid %r had responses already' % (xid,))
        if xid in connection.http_responsebuffer:
            rbuffer = connection.http_responsebuffer[xid]
            if len(rbuffer) > 0:
                last = rbuffer[-1]
                if last.status[0] != b'1'[0]:
                    if response.status[0] != b'1'[0]:
                        raise HttpProtocolException('Cannot response with a non-1xx status code more than once')
                    else:
                        rbuffer.insert(-1, response)
                else:
                    rbuffer.append(response)
            else:
                rbuffer.append(response)
                if xid == connection.http_responsexid:
                    return HttpStateChange(connection, connection.connmark, HttpStateChange.NEXTOUTPUT)
        else:
            connection.http_responsebuffer[xid] = [response]
            if xid == connection.http_responsexid:
                return HttpStateChange(connection, connection.connmark, HttpStateChange.NEXTOUTPUT)
        return None
    def startResponse(self, connection, xid, status, headers, outputstream, disabledeflate = False):
        """
        Start to response to a request with the specified xid on the connection, with status code
        and headers. The output stream is used to output the response body.
        """
        resp = self.createResponse(connection, xid, status, headers, outputstream)
        resp.disabledeflate = disabledeflate
        event = self.responseTo(connection, xid, resp)
        if event is not None:
            connection.scheduler.emergesend(event)
        return resp
    def _createrequestheaders(self, connection, host, method, headers, stream = None, keepalive = True):
        defaultheaders = self.defaultrequestheaders
        existingHeaders = set(normalizeHeader(k) for k,_ in headers)
        if connection.http_localversion < '1.1' or not keepalive:
            newheaders = [(b'Connection', b'Close')]
        else:
            newheaders = [(b'Connection', b'Keep-Alive')]
        existingHeaders.add(b'connection')
        if host is not None:
            newheaders.append((b'Host', host))
        existingHeaders.add(b'host')
        if connection.http_localversion >= '1.1':
            if self.allowtransfercompress and b'te' not in existingHeaders:
                existingHeaders.add(b'te')
                newheaders.append((b'TE', b'deflate'))
        if b'content-length' not in existingHeaders and method != b'CONNECT':
            if stream is None:
                size = 0
            else:
                try:
                    size = len(stream)
                except:
                    size = None
            if size is None:
                pass
            elif size == 0:
                if method in (b'PUT', b'POST') or stream is not None:
                    # If you give a MemoryStream(b'') to a GET request
                    # there will be a content-length header, which may cause problem
                    # for some servers
                    newheaders.append((b'Content-Length', b'0'))
            else:
                newheaders.append((b'Content-Length', str(size).encode('ascii')))
            existingHeaders.add(b'content-length')
        for k,v in defaultheaders:
            if normalizeHeader(k) not in existingHeaders:
                newheaders.append((k, v))
        newheaders.extend((k,v) for k,v in headers if normalizeHeader(k) not in (b'connection', b'transfer-encoding', b'host', b'expect'))
        return newheaders
    def sendRequest(self, connection, host, path = b'/', method = b'GET', headers = [], stream = None, keepalive = True):
        '''
        If you do not provide a content-length header, the stream will be transfer-encoded with chunked, and it is not
        always acceptable by servers.
        
        You may provide a MemoryStream, and it will provide a content-length header automatically
        
        Return xid
        '''
        method = method.upper()
        headers = self._createrequestheaders(connection, host, method, headers, stream, keepalive)
        if not connection.connected or connection.http_requestbuffer is None:
            raise HttpConnectionClosedException('Connection is closed')
        notify = not connection.http_requestbuffer
        connection.http_requestbuffer.append((host, path, method, headers, stream, keepalive))
        if notify:
            connection.scheduler.emergesend(HttpStateChange(connection, connection.connmark, HttpStateChange.NEXTOUTPUT))
        return connection.xid + len(connection.http_requestbuffer) - 1
    def responsematcher(self, connection, xid, isfinal = None, iserror = None):
        """
        Create an event matcher to match the response
        """
        return HttpResponseEvent.createMatcher(connection, connection.connmark, xid, isfinal, iserror)
    def statematcher(self, connection, state = HttpConnectionStateEvent.CLIENT_CLOSE, currentconn = True):
        """
        Create an event matcher to match the connection state
        """
        return HttpConnectionStateEvent.createMatcher(state, connection, connection.connmark if currentconn else None)
    def requestwithresponse(self, container, connection, host, path = b'/', method = b'GET', headers = [], stream = None, keepalive = True):
        """
        Send a HTTP request, and wait for the response. The last (usually wanted) response is stored in
        `container.http_finalresponse`. There may be multiple responses (1xx) for this request, they are
        stored in `container.http_responses`
        """
        xid = self.sendRequest(connection, host, path, method, headers, stream, keepalive)
        resp = self.responsematcher(connection, xid)
        stat = self.statematcher(connection)
        resps = []
        while True:
            yield (resp, stat)
            if container.matcher is stat:
                raise HttpConnectionClosedException('Connection closed before response received')
            r = container.event
            resps.append(r)
            if r.isfinal:
                container.http_finalresponse = r
                break
        container.http_responses = resps
    def parse(self, connection, data, laststart):
        events = []
        if hasattr(data, 'tobytes'):
            data = data.tobytes()
        else:
            data = data[:]

        if connection.http_parsestage == 'end':
            # Ignore more data
            return (events, 0)
        if connection.http_remoteversion is None and not self.server and \
                connection.http_parsestage == 'headline' and self.allownoheader:
            # Try to find out if there is a status line as early as possible
            # len('HTTP/1.1 ') == 9 
            if len(data) <= 9:
                hasstatus = statuscheckshort.match(data)
            else:
                hasstatus = statuscheck.match(data)
            if not hasstatus:
                connection.http_parsestage = 'body'
                connection.http_remoteversion = '0.9'
                connection.http_keepalive = False
                connection.http_contentlength = None
                connection.http_deflate = False
                connection.http_chunked = False
                connection.http_dataread = 0
                responsestream = Stream()
                responsestream.trailers = []
                events.append(HttpResponseEvent(connection,
                                             connection.connmark,
                                             connection.http_responsexid,
                                             True,
                                             False,
                                             statuscode = 200,
                                             status = b'200 OK',
                                             statustext = b'OK',
                                             headers = [],
                                             headerdict = {},
                                             setcookies = [],
                                             stream = responsestream
                                             ))
                connection.http_currentstream = responsestream
                connection.http_responsexid += 1
        stage = connection.http_parsestage
        start = 0
        end = len(data)
        def httpfail(code = 400):
            if self.debugging:
                self._logger.debug('HTTP Protocol violation detected')
            if not self.server or self.fastfail:
                events.append(ConnectionControlEvent(connection, ConnectionControlEvent.RESET, True, connection.connmark))
            else:
                connection.http_keepalive = False
                event = self.responseTo(connection, connection.xid,
                            self.createErrorResponse(connection, connection.xid, code))
                if connection.xid == connection.http_responsexid:
                    events.append(HttpStateChange(connection, connection.connmark, HttpStateChange.NEXTINPUT))
                connection.http_reqteinfo.append((b'GET', False, False, False))
                connection.xid += 1
                connection.http_idle = True
                if event is not None:
                    events.append(event)
        def stopstream(stoptype = StreamDataEvent.STREAM_ERROR):
            events.append(StreamDataEvent(connection.http_currentstream,
                          stoptype,
                          data = b''))
            if not hasattr(connection.http_currentstream, 'trailers'):
                connection.http_currentstream.trailers = []
                events.append(HttpTrailersReceived(connection.http_currentstream))
            connection.http_currentstream.writeclosed = True
            connection.http_currentstream = None
            connection.http_keepalive = False
            connection.http_idle = True
            if not self.server:
                events.append(ConnectionControlEvent(connection, ConnectionControlEvent.SHUTDOWN, False, connection.connmark))                            
                events.append(HttpResponseEndEvent(connection, connection.connmark,
                                                   connection.http_responsexid - 1,
                                                   False))
        while start < end:
            if stage == 'end':
                start = end
            elif stage == 'headers':
                ls = data.find(b'\r\n\r\n', start)
                if ls < 0:
                    break
                header_lines = data[start:ls].split(b'\r\n')
                start = ls + 4
                if self.headercountlimit is not None and len(header_lines) > self.headercountlimit:
                    self._logger.info('Bad header detected: header count limit exceeded')
                    stage = 'end'
                    httpfail(431)
                    continue
                http_headers = []
                for line in header_lines:
                    key, sep, value = line.partition(b':')
                    if not sep or not token_r.match(key):
                        # Bad header
                        self._logger.info('Bad header detected: %r', line)
                        stage = 'end'
                        httpfail(400)
                        break
                    else:
                        http_headers.append((key, value.strip()))
                if stage == 'end':
                    continue
                connection.http_headers = http_headers
                # An empty line indicates end of headers
                # Processing headers
                headerdict = {}
                setcookies = []
                for k,v in http_headers:
                    kn = normalizeHeader(k)
                    if kn == b'set-cookie':
                        # RFC said set-cookie is a special case
                        # it may appear multiple times and cannot be joined
                        setcookies.append(v)
                    elif kn in headerdict:
                        # RFC said we can join multiple headers with ,
                        # without changing the meaning
                        headerdict[kn] = headerdict[kn] + b', ' + v
                    else:
                        headerdict[kn] = v
                if connection.http_localversion >= '1.1':
                    connection_options = [normalizeHeader(k.strip()) for k in headerdict.get(b'connection', b'').split(b',') if k.strip() != b'']
                    if connection.http_remoteversion < '1.1':
                        connection.http_keepalive = b'keep-alive' in connection_options
                        #connection.http_keepalive = False
                    else:
                        connection.http_keepalive = b'close' not in connection_options
                else:
                    connection.http_keepalive = False
                # reqteinfo is in format: (method, acceptTrailers, acceptDeflate, acceptChunk) 
                if self.server:
                    if connection.http_method == b'HEAD':
                        connection.http_reqteinfo.append((connection.http_method, False, False, False))
                    elif connection.http_remoteversion >= '1.1':
                        transfer_options = tuple(normalizeHeader(k.strip()) for k in headerdict.get(b'te', b'').split(b',') if k.strip() != b'')
                        transfer_trailers = b'trailers' in transfer_options
                        transfer_te = tuple(k.split(b';',2)[0].strip() for k in transfer_options if k != b'trailers')
                        transfer_deflate = b'deflate' in transfer_te
                        connection.http_reqteinfo.append((connection.http_method, transfer_trailers, transfer_deflate, True))
                    else:
                        connection.http_reqteinfo.append((connection.http_method, False, False, False))
                # Determine message body
                if self.server:
                    stage = 'body'
                    if connection.http_method == b'CONNECT':
                        # A tunnel
                        connection.http_contentlength = None
                        connection.http_chunked = False
                        connection.http_deflate = False
                        connection.http_keepalive = False
                    elif connection.http_remoteversion >= '1.1' and b'transfer-encoding' in headerdict:
                        connection.http_contentlength = None
                        transferencoding = tuple(k.strip().lower() for k in headerdict[b'transfer-encoding'].split(b',') if k.strip() != b'')
                        if transferencoding and transferencoding[-1] == b'chunked':
                            connection.http_chunked = True
                            transferencoding = transferencoding[:-1]
                        # For security reason, we do not accept data compressed more than once
                        # Do not allow a chunked in chunked
                        if not transferencoding:
                            connection.http_deflate = False
                            if not connection.http_chunked:
                                connection.http_keepalive = False
                        elif len(transferencoding) > 1 or transferencoding[0] == b'chunked':
                            if self.debugging:
                                self._logger.debug('Unacceptable transfer encoding: %r', headerdict[b'transfer-encoding'])
                            stage = 'end'
                            httpfail(400)
                        elif transferencoding[0] != b'deflate':
                            if self.debugging:
                                self._logger.debug('Unacceptable transfer encoding: %r', headerdict[b'transfer-encoding'])
                            # We only accept deflate as the compress format
                            stage = 'end'
                            httpfail(501)
                        else:
                            connection.http_deflate = True
                            if not connection.http_chunked:
                                connection.http_keepalive = False
                    elif b'content-length' in headerdict:
                        try:
                            connection.http_contentlength = int(headerdict[b'content-length'])
                        except:
                            # Illegal content-length
                            if self.debugging:
                                self._logger.debug('Illegal content length: %r', headerdict[b'content-length'])
                            stage = 'end'
                            httpfail(400)
                        else:
                            if connection.http_contentlength < 0:
                                # Illegal content-length
                                if self.debugging:
                                    self._logger.debug('Illegal content length: %r', headerdict[b'content-length'])
                                stage = 'end'
                                httpfail(400)
                            elif connection.http_contentlength == 0:
                                # Empty body
                                connection.http_contentlength = 0
                            else:
                                connection.http_deflate = False
                                connection.http_chunked = False
                    else:
                        # Empty body
                        connection.http_contentlength = 0
                else:
                    if connection.http_statuscode >= 100 and connection.http_statuscode < 200:
                        # 1xx response has no body, and is not a final response
                        connection.http_contentlength = 0
                    else:
                        teinfo = connection.http_reqteinfo.popleft()
                        if teinfo[0] == b'HEAD':
                            # HEAD response has no body
                            connection.http_chunked = False
                            connection.http_deflate = False
                            connection.http_contentlength = 0
                        elif teinfo[0] == b'CONNECT':
                            connection.http_contentlength = None
                            connection.http_chunked = False
                            connection.http_deflate = False
                        elif connection.http_statuscode == 304 or connection.http_statuscode == 204:
                            # 204 and 304 response has no body
                            connection.http_chunked = False
                            connection.http_deflate = False
                            connection.http_contentlength = 0
                        elif connection.http_remoteversion >= '1.1' and b'transfer-encoding' in headerdict:
                            connection.http_contentlength = None
                            transferencoding = tuple(k.strip().lower() for k in headerdict[b'transfer-encoding'].split(b',') if k.strip() != b'')
                            if transferencoding and transferencoding[-1] == b'chunked':
                                connection.http_chunked = True
                                transferencoding = transferencoding[:-1]
                            # For security reason, we do not accept data compressed more than once
                            # Do not allow a chunked in chunked
                            if not transferencoding:
                                connection.http_deflate = False
                                if not connection.http_chunked:
                                    connection.http_keepalive = False
                            elif len(transferencoding) > 1 or transferencoding[0] == b'chunked':
                                if self.debugging:
                                    self._logger.debug('Unacceptable transfer encoding: %r', headerdict[b'transfer-encoding'])
                                stage = 'end'
                                httpfail()
                            elif transferencoding[0] != b'deflate':
                                # We only accept deflate as the compress format
                                if self.debugging:
                                    self._logger.debug('Unacceptable transfer encoding: %r', headerdict[b'transfer-encoding'])
                                stage = 'end'
                                httpfail()
                            else:
                                connection.http_deflate = True
                                if not connection.http_chunked:
                                    connection.http_keepalive = False
                        elif b'content-length' in headerdict:
                            try:
                                connection.http_contentlength = int(headerdict[b'content-length'])
                            except:
                                # Illegal content-length
                                stage = 'end'
                                httpfail()
                            else:
                                if connection.http_contentlength < 0:
                                    # Illegal content-length
                                    httpfail()
                                elif connection.http_contentlength == 0:
                                    # Empty body
                                    connection.http_contentlength = 0
                                else:
                                    connection.http_deflate = False
                                    connection.http_chunked = False
                        else:
                            # A undetermined-length body
                            connection.http_contentlength = None
                            connection.http_deflate = False
                            connection.http_chunked = False
                            connection.http_keepalive = False
                expect_100 = False
                if stage != 'end' and self.server:
                    if b'expect' in headerdict:
                        if headerdict[b'expect'].lower() != b'100-continue':
                            if self.debugging:
                                self._logger.debug('Unacceptable expect: %r', headerdict[b'expect'])
                            httpfail(417)
                        else:
                            expect_100 = True
                    else:
                        expect_100 = False
                if stage != 'end':
                    if connection.http_contentlength == 0:
                        inputstream = MemoryStream(b'')
                        if connection.http_keepalive:
                            stage = 'headline'
                        else:
                            stage = 'end'
                        connection.http_idle = True
                        inputstream.trailers = []
                    else:
                        if expect_100:
                            inputstream = self._HttpContinueStream(self, connection, connection.xid)
                        else:
                            inputstream = Stream()
                        if connection.http_deflate:
                            connection.http_deflateobj = zlib.decompressobj()
                        if connection.http_chunked:
                            stage = 'chunkedheader'
                        else:
                            stage = 'body'
                            inputstream.trailers = []
                            connection.http_dataread = 0
                        connection.http_currentstream = inputstream
                    if self.server:
                        self._logger.info('Request received from connection %r, xid = %d: %r', connection, connection.xid,
                                          connection.http_method + b' ' + connection.http_path + b' HTTP/' + connection.http_remoteversion.encode('ascii'))
                        r = HttpRequestEvent(headerdict.get(b'host', b''),
                                                       connection.http_path,
                                                       connection.http_method,
                                                       connection,
                                                       connection.connmark,
                                                       connection.xid,
                                                       self,
                                                       headers = connection.http_headers,
                                                       headerdict = headerdict,
                                                       setcookies = setcookies,
                                                       stream = inputstream
                                                       )
                        if connection.xid >= connection.http_responsexid + self.pipelinelimit:
                            connection.http_requestbuffer.append(r)
                        else:
                            events.append(r)
                        connection.xid += 1
                    else:
                        self._logger.info('Response received from connection %r, xid = %d: %r', connection, connection.http_responsexid,
                                          b'HTTP/' + connection.http_remoteversion.encode('utf-8') + b' ' + connection.http_status)
                        events.append(HttpResponseEvent(connection,
                                                     connection.connmark,
                                                     connection.http_responsexid,
                                                     connection.http_statuscode < 100 or connection.http_statuscode >= 200,
                                                     connection.http_statuscode >= 400,
                                                     statuscode = connection.http_statuscode,
                                                     status = connection.http_status,
                                                     statustext = connection.http_statustext,
                                                     headers = connection.http_headers,
                                                     headerdict = headerdict,
                                                     setcookies = setcookies,
                                                     stream = inputstream
                                                     ))
                        if connection.http_statuscode < 100 or connection.http_statuscode >= 200:
                            connection.http_responsexid += 1
                            if stage == 'headline':
                                events.append(HttpResponseEndEvent(connection, connection.connmark,
                                                                   connection.http_responsexid - 1,
                                                                   True))
                            elif stage == 'end':
                                events.append(HttpResponseEndEvent(connection, connection.connmark,
                                                                   connection.http_responsexid - 1,
                                                                   False))
            elif stage == 'headline' or stage == 'chunkedheader' or stage == 'trailers':
                ls = data.find(b'\r\n', start)
                if ls < 0:
                    break
                le = ls + 2
                if stage == 'headline':
                    if start == ls:
                        # RFC said that we should ignore at least one empty line
                        # to achieve robustness
                        pass
                    elif self.server:
                        match = requestline.match(data[start:ls])
                        if match is None:
                            # Bad request
                            if connection.http_remoteversion is None:
                                connection.http_remoteversion = '1.0'
                            stage = 'end'
                            if self.debugging:
                                if ls - start > 200:
                                    self._logger.info('Invalid request from connection %r: %r...', connection, data[start:start + 200])
                                else:
                                    self._logger.info('Invalid request from connection %r: %r', connection, data[start:ls])
                            httpfail(400)
                        else:
                            connection.http_remoteversion = str(match.group(3).decode('ascii'))
                            connection.http_path = match.group(2)
                            connection.http_method = match.group(1).upper()
                            connection.http_headers = []
                            stage = 'headers'
                            if self.debugging:
                                if ls - start > 200:
                                    self._logger.debug('First line of request received from connection %r: %r...', connection, data[start:start + 200])
                                else:
                                    self._logger.debug('First line of request received from connection %r: %r', connection, data[start:ls])
                            connection.http_idle = False
                            if connection.xid == connection.http_responsexid:
                                events.append(HttpStateChange(connection, connection.connmark, HttpStateChange.NEXTINPUT))
                    else:
                        match = statusline.match(data[start:ls])
                        if match is None:
                            if self.debugging:
                                self._logger.info('Bad headline detected: %r', data[start:ls])
                            # Bad response
                            stage = 'end'
                            connection.http_remoteversion = '1.0'
                            httpfail()
                        else:
                            connection.http_remoteversion = str(match.group(1).decode('ascii'))
                            connection.http_status = match.group(2) + b' ' + match.group(3)
                            connection.http_statuscode = int(match.group(2))
                            connection.http_statustext = match.group(3)
                            connection.http_headers = []
                            stage = 'headers'
                            if self.debugging:
                                if ls - start > 200:
                                    self._logger.debug('First line of response received from connection %r: %r...', connection, data[start:start + 200])
                                else:
                                    self._logger.debug('First line of response received from connection %r: %r', connection, data[start:ls])
                            connection.http_idle = False
                elif stage == 'chunkedheader':
                    chunked_header = data[start:ls]
                    chunked_header, _, _ = chunked_header.partition(b';')
                    if self.chunkednumberlimit is not None and len(chunked_header) > self.chunkednumberlimit:
                        stage = 'end'
                        stopstream()
                    else:
                        try:
                            chunked_size = int(chunked_header, 16)
                        except Exception:
                            stage = 'end'
                            stopstream(StreamDataEvent.STREAM_ERROR)
                        else:
                            connection.http_chunkedsize = chunked_size
                            if chunked_size == 0:
                                # Last chunked
                                events.append(StreamDataEvent(connection.http_currentstream,
                                                              StreamDataEvent.STREAM_EOF,
                                                              data = b''))
                                connection.http_currentstream.writeclosed = True
                                connection.http_trailers = []
                                stage = 'trailers'
                            else:
                                connection.http_dataread = 0
                                stage = 'body'
                else:
                    # Tailers
                    if ls == start:
                        # End of chunked
                        connection.http_currentstream.trailers = connection.http_trailers
                        events.append(HttpTrailersReceived(connection.http_currentstream))
                        connection.http_currentstream = None
                        connection.http_idle = True
                        if connection.http_keepalive:
                            stage = 'headline'
                            if not self.server:
                                if connection.http_statuscode < 100 or connection.http_statuscode >= 200:
                                    events.append(HttpResponseEndEvent(connection, connection.connmark,
                                                                       connection.http_responsexid - 1,
                                                                       True))
                        else:
                            stage = 'end'
                            if not self.server:
                                # Close connection
                                events.append(ConnectionControlEvent(connection, ConnectionControlEvent.SHUTDOWN, False, connection.connmark))
                                if connection.http_statuscode < 100 or connection.http_statuscode >= 200:
                                    events.append(HttpResponseEndEvent(connection, connection.connmark,
                                                                       connection.http_responsexid - 1,
                                                                       True))
                    else:
                        match = headerline.match(data[start:ls])
                        if match is None:
                            # Bad header
                            stage = 'end'
                            connection.http_keepalive = False
                            # Ignore all trailers
                            connection.http_currentstream.trailers = []
                            events.append(HttpTrailersReceived(connection.http_currentstream))
                            connection.http_currentstream = None
                        else:
                            connection.http_trailers.append((match.group(1), match.group(2)))
                            if self.headercountlimit is not None and len(connection.http_trailers) > self.headercountlimit:
                                stage = 'end'
                                connection.http_keepalive = False
                                # Ignore all trailers
                                connection.http_currentstream.trailers = []
                                events.append(HttpTrailersReceived(connection.http_currentstream))
                                connection.http_currentstream = None
                                connection.http_idle = True
                start = le
            elif stage == 'chunkedtail':
                # Always \r\n
                if start + 2 > end:
                    break
                if not data[start:start+2] == b'\r\n':
                    stage = 'end'
                    stopstream()
                else:
                    start += 2
                    stage = 'chunkedheader'
            else:
                if stage != 'body':
                    raise ValueError('stage == ' + stage)
                # Body
                if connection.http_chunked:
                    totalsize = connection.http_chunkedsize
                else:
                    totalsize = connection.http_contentlength
                if getattr(connection, 'http_dataread', None) is None:
                    raise ValueError('Invalid read on connection: ' + repr(connection) + ',' + repr(totalsize) + ',' + repr(getattr(connection, 'http_dataread', None)))
                if totalsize is not None:
                    resumesize = totalsize - connection.http_dataread
                if totalsize is not None and resumesize <= end - start:
                    readdata = data[start: start + resumesize]
                    start += resumesize
                    connection.http_dataread += resumesize
                    if connection.http_chunked:
                        stage = 'chunkedtail'
                        eof = False
                    elif connection.http_keepalive:
                        stage = 'headline'
                        eof = True
                        connection.http_idle = True
                        if not self.server:
                            if connection.http_statuscode < 100 or connection.http_statuscode >= 200:
                                events.append(HttpResponseEndEvent(connection, connection.connmark,
                                                                   connection.http_responsexid - 1,
                                                                   True))
                    else:
                        stage = 'end'
                        eof = True
                        connection.http_idle = True
                        if not self.server:
                            if connection.http_statuscode < 100 or connection.http_statuscode >= 200:
                                events.append(HttpResponseEndEvent(connection, connection.connmark,
                                                                   connection.http_responsexid - 1,
                                                                   False))
                    if connection.http_deflate:
                        try:
                            readdata = connection.http_deflateobj.decompress(readdata)
                            readdata += connection.http_deflateobj.flush()
                        except zlib.error:
                            stage = 'end'
                            stopstream()
                            continue
                    if readdata or eof:
                        events.append(StreamDataEvent(connection.http_currentstream,
                                                      StreamDataEvent.STREAM_EOF if eof else StreamDataEvent.STREAM_DATA,
                                                      data = readdata))
                    if eof:
                        connection.http_currentstream.writeclosed = True
                        connection.http_currentstream = None
                        if not self.server and stage == 'end':
                            events.append(ConnectionControlEvent(connection, ConnectionControlEvent.SHUTDOWN, False, connection.connmark))                            
                else:
                    readdata = data[start:]
                    if connection.http_deflate:
                        try:
                            readdata = connection.http_deflateobj.decompress(readdata)
                        except zlib.error:
                            stage = 'end'
                            stopstream()
                            continue
                    if readdata:
                        events.append(StreamDataEvent(connection.http_currentstream,
                                                      StreamDataEvent.STREAM_DATA,
                                                      data = readdata))
                    connection.http_dataread += end - start
                    start = end
        if stage == 'headline' and self.headlinelimit is not None and end - start > self.headlinelimit:
            stage = 'end'
            start = end
            if self.debugging:
                self._logger.debug('Headline exceeds limit on connection %r: %r...', connection, data[start:start + 100])
            httpfail(400)
        if stage == 'headers' and self.headerlimit is not None and end - start > self.headerlimit:
            stage = 'end'
            start = end
            if self.debugging:
                self._logger.debug('A single header exceeds limit on connection %r: %r...', connection, data[start:start + 100])
            httpfail(431)
        if stage == 'chunkedheader' and self.chunkedheaderlimit is not None and end - start > self.chunkedheaderlimit:
            stage = 'end'
            start = end
            if self.debugging:
                self._logger.debug('A chunked header exceeds limit on connection %r: %r...', connection, data[start:start + 100])
            stopstream()
        if stage == 'body' and laststart == end:
            stage = 'end'
            if not connection.http_chunked and connection.http_contentlength is None:
                stopstream(StreamDataEvent.STREAM_EOF)
            else:
                stopstream(StreamDataEvent.STREAM_ERROR)
        elif laststart == end:
            # Remote write-close.
            if not self.server:
                # server closed the connection, we should also close the connection
                httpfail()
        connection.http_parsestage = stage
        return (events, end - start)
    def beforelisten(self, tcpserver, newsock):
        try:
            tcpserver.scheduler.queue['read'].addSubQueue(5, PollEvent.createMatcher(newsock.fileno(), PollEvent.READ_READY), tcpserver)
        except:
            pass
        if False:
            yield

    def serverfinal(self, tcpserver):
        try:
            for m in tcpserver.syscall(syscall_clearremovequeue(tcpserver.scheduler.queue['read'], tcpserver)):
                yield m
        except:
            pass
    def notconnected(self, connection):
        for m in connection.waitForSend(HttpConnectionStateEvent(HttpConnectionStateEvent.CLIENT_NOTCONNECTED,
                                                                 connection,
                                                                 connection.connmark,
                                                                 self)):
            yield m
    def final(self, connection):
        if hasattr(connection, 'http_sender') and connection.http_sender:
            connection.http_sender.close()
        for m in Protocol.final(self, connection):
            yield m
    def waitForResponseEnd(self, container, connection, connmark, xid):
        if not connection.connected or connection.connmark != connmark:
            container.retvalue = False
        elif connection.http_responsexid > xid + 1:
            container.retvalue = True
        elif connection.http_responsexid == xid + 1 and connection.http_parsestage in ('end', 'headline', 'headers'):
            container.retvalue = connection.http_parsestage != 'end'
        else:
            re = HttpResponseEndEvent.createMatcher(connection, connmark, xid)
            rc = self.statematcher(connection)
            yield (re, rc)
            if container.matcher is re:
                container.retvalue = container.event.keepalive
            else:
                container.retvalue = False



'''
Created on 2015/12/25

:author: hubo
'''
from vlcp.config import defaultconfig
from vlcp.protocol.jsonrpc import JsonRPC
from vlcp.service.connection.tcpserver import TcpServerBase
from vlcp.protocol.ovsdb import OVSDB

@defaultconfig
class JsonRPCServer(TcpServerBase):
    '''
    Create JsonRPC server on specified URLs, vHosts are supported.
    '''
    # Enable connection management
    _default_connmanage = True
    def __init__(self, server):
        TcpServerBase.__init__(self, server, JsonRPC)
    def _createprotocol(self, config):
        if getattr(config, 'ovsdb', False):
            return OVSDB()
        else:
            return JsonRPC()



'''
Created on 2016/4/11

:author: hubo
'''
from vlcp.event.runnable import RoutineContainer
from vlcp.server.module import callAPI
from uuid import uuid1
from vlcp.event.event import Event, withIndices
from vlcp.utils.dataobject import multiwaitif
from vlcp.protocol.openflow.openflow import OpenflowErrorResultException
from namedstruct.namedstruct import dump
import json
import logging

@withIndices('updater', 'type')
class FlowUpdaterNotification(Event):
    STARTWALK = 'startwalk'
    DATAUPDATED = 'dataupdated'
    FLOWUPDATE = 'flowupdate'

class FlowUpdater(RoutineContainer):
    def __init__(self, connection, initialkeys, requestid = None, logger = None):
        RoutineContainer.__init__(self, connection.scheduler)
        self._initialkeys = initialkeys
        self._connection = connection
        self._walkerdict = {}
        self._savedkeys = ()
        self._savedresult = []
        self._updatedset = set()
        self._updatedset2 = set()
        if not logger:
            self._logger = logging.getLogger(__name__ + '.FlowUpdater')
        else:
            self._logger = logger
        if requestid is None:
            self._requstid = str(uuid1())
        else:
            self._requstid = requestid
        self._dataupdateroutine = None
        self._flowupdateroutine = None
    def reset_initialkeys(self, keys, values):
        pass
    def walkcomplete(self, keys, values):
        if False:
            yield
    def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        if False:
            yield
    def shouldupdate(self, newvalues, updatedvalues):
        return True
    def restart_walk(self):
        self._restartwalk = True
        for m in self.waitForSend(FlowUpdaterNotification(self, FlowUpdaterNotification.STARTWALK)):
            yield m
    def _dataobject_update_detect(self):
        _initialkeys = set(self._initialkeys)
        def expr(newvalues, updatedvalues):
            if any(v.getkey() in _initialkeys for v in updatedvalues if v is not None):
                return True
            else:
                return self.shouldupdate(newvalues, updatedvalues)
        while True:
            for m in multiwaitif(self._savedresult, self, expr, True):
                yield m
            updatedvalues, _ = self.retvalue
            if not self._updatedset:
                self.scheduler.emergesend(FlowUpdaterNotification(self, FlowUpdaterNotification.DATAUPDATED))
            self._updatedset.update(updatedvalues)
    def updateobjects(self, updatedvalues):
        if not self._updatedset:
            self.scheduler.emergesend(FlowUpdaterNotification(self, FlowUpdaterNotification.DATAUPDATED))
        self._updatedset.update(set(updatedvalues).intersection(self._savedresult))
    def _flowupdater(self):
        lastresult = set(v for v in self._savedresult if v is not None and not v.isdeleted())
        flowupdate = FlowUpdaterNotification.createMatcher(self, FlowUpdaterNotification.FLOWUPDATE)
        while True:
            currentresult = set(v for v in self._savedresult if v is not None and not v.isdeleted())
            additems = currentresult.difference(lastresult)
            removeitems = lastresult.difference(currentresult)
            updateditems = set(self._updatedset2).intersection(currentresult)
            updateditems.difference_update(removeitems)
            updateditems.difference_update(additems)
            self._updatedset2.clear()
            lastresult = currentresult
            if not additems and not removeitems and not updateditems:
                yield (flowupdate,)
                continue
            for m in self.updateflow(self._connection, additems, removeitems, updateditems):
                yield m
                
    def main(self):
        try:
            lastkeys = set()
            dataupdate = FlowUpdaterNotification.createMatcher(self, FlowUpdaterNotification.DATAUPDATED)
            startwalk = FlowUpdaterNotification.createMatcher(self, FlowUpdaterNotification.STARTWALK)
            self.subroutine(self._flowupdater(), False, '_flowupdateroutine')
            presave_update = set()
            while True:
                self._restartwalk = False
                presave_update.clear()
                presave_update.update(self._updatedset)
                self._updatedset.clear()
                _initialkeys = set(self._initialkeys)
                for m in callAPI(self, 'objectdb', 'walk', {'keys': self._initialkeys, 'walkerdict': self._walkerdict,
                                                            'requestid': self._requstid}):
                    yield m
                if self._updatedset:
                    if any(v.getkey() in _initialkeys for v in self._updatedset):
                        continue
                lastkeys = set(self._savedkeys)
                self._savedkeys, self._savedresult = self.retvalue
                removekeys = tuple(lastkeys.difference(self._savedkeys))
                self.reset_initialkeys(self._savedkeys, self._savedresult)
                _initialkeys = set(self._initialkeys)
                if self._dataupdateroutine:
                    self.terminate(self._dataupdateroutine)
                self.subroutine(self._dataobject_update_detect(), False, "_dataupdateroutine")
                self._updatedset.update(v for v in presave_update)
                if removekeys:
                    for m in callAPI(self, 'objectdb', 'munwatch', {'keys': removekeys,
                                                                    'requestid': self._requstid}):
                        yield m
                for m in self.walkcomplete(self._savedkeys, self._savedresult):
                    yield m
                self._updatedset2.update(self._updatedset)
                self._updatedset.clear()
                for m in self.waitForSend(FlowUpdaterNotification(self, FlowUpdaterNotification.FLOWUPDATE)):
                    yield m
                while not self._restartwalk:
                    if self._updatedset:
                        if any(v.getkey() in _initialkeys for v in self._updatedset):
                            break
                        else:
                            self._updatedset2.update(self._updatedset)
                            self._updatedset.clear()
                            self.scheduler.emergesend(FlowUpdaterNotification(self, FlowUpdaterNotification.FLOWUPDATE))
                    yield (dataupdate, startwalk)
        except GeneratorExit:
            raise
        except Exception:
            self._logger.exception("Flow updater %r stops update by an exception, conn = %r", self, self._connection)
            raise
        finally:
            self.subroutine(callAPI(self, 'objectdb', 'munwatch', {'keys': self._savedkeys,
                                                                   'requestid': self._requstid}))
            if self._flowupdateroutine:
                self.terminate(self._flowupdateroutine)
                self._flowupdateroutine = None
            if self._dataupdateroutine:
                self.terminate(self._dataupdateroutine)
                self._dataupdateroutine = None
    def execute_commands(self, conn, cmds):
        if cmds:
            try:
                for m in conn.protocol.batch(cmds, conn, self):
                    yield m
            except OpenflowErrorResultException:
                self._logger.warning("Some Openflow commands return error result on connection %r, will ignore and continue.\n"
                                             "Details:\n%s", conn,
                                             "\n".join("REQUEST = \n%s\nERRORS = \n%s\n" % (json.dumps(dump(k), indent=2),
                                                                                            json.dumps(dump(v), indent=2))
                                                       for k,v in self.openflow_replydict.items()))



'''
Created on 2015/06/02

:author: hubo
'''
from __future__ import print_function, absolute_import, division 

try:
    from vlcp_event_cython.pqueue import *
except:

    from collections import deque
    from .matchtree import MatchTree
    from .event import Event, withIndices
    from bisect import bisect_left
    from heapq import heappush, heappop

    @withIndices('queue')
    class QueueCanWriteEvent(Event):
        pass

    @withIndices('queue')
    class QueueIsEmptyEvent(Event):
        pass

    @withIndices('newonly', 'firstonly')
    class AutoClassQueueCanWriteEvent(QueueCanWriteEvent):
        pass

    class CBQueue(object):
        '''
        A multi-queue model with priority and balance.
        When first created, there is a default queue with priority 0. More sub-queues maybe created with addSubQueue.
        Each sub-queue is a CBQueue which accepts more sub-queues. Sub-queues are considered as black-box to the outer parent.
        '''
        class FifoQueue(object):
            '''
            A wrapper for a FIFO queue
            '''
            def __init__(self, parent = None, maxlength = None):
                self.queue = deque()
                self.parent = parent
                self.maxlength = maxlength
                self.blocked = False
                if self.maxlength is not None and self.maxlength <= 0:
                    self.maxlength = 1
                self.isWaited = False
            def append(self, value, force = False):
                if not force and not self.canAppend():
                    self.isWaited = True
                    return QueueCanWriteEvent.createMatcher(self)
                if self.parent is not None:
                    m = self.parent.notifyAppend(self, force)
                    if m is not None:
                        return m
                self.queue.append(value)
                return None
            def canAppend(self):
                return self.maxlength is None or len(self.queue) < self.maxlength
            def canPop(self):
                return self.queue and not self.blocked
            def pop(self):
                ret = self._pop()
                if self.parent is not None:
                    pr = self.parent.notifyPop(self)
                    ret[1].extend(pr[0])
                    ret[2].extend(pr[1])
                return ret
            def _pop(self):
                if self.blocked:
                    raise IndexError('pop from a blocked queue')
                ret = self.queue.popleft()
                if self.isWaited and self.canAppend():
                    self.isWaited = False
                    return (ret, [QueueCanWriteEvent(self)], [])
                else:
                    return (ret, [], [])
            def clear(self):
                l = len(self)
                ret = self._clear()
                if self.parent is not None:
                    pr = self.parent.notifyPop(self, l)
                    ret[0].extend(pr[0])
                    ret[1].extend(pr[1])
                return ret
            def _clear(self):
                if self.blocked:
                    self.unblockall()
                self.queue.clear()
                if self.isWaited and self.canAppend():
                    self.isWaited = False
                    return ([QueueCanWriteEvent(self)], [])
                else:
                    return ([], [])
            def __len__(self):
                return len(self.queue)
            def block(self, value):
                if self.parent is not None:
                    self.parent.notifyAppend(self, True)
                self.queue.appendleft(value)
                if not self.blocked:
                    self.blocked = True
                    if self.parent is not None:
                        self.parent.notifyBlock(self, True)
            def unblock(self, value):
                if self.blocked:
                    self.blocked = False
                    if self.parent is not None:
                        self.parent.notifyBlock(self, False)
            def unblockall(self):
                if self.blocked:
                    self.blocked = False
                    if self.parent is not None:
                        self.parent.notifyBlock(self, False)
        class PriorityQueue(object):
            '''
            A queue with inner built priority. Event must have a "priority" property to use with this type of queue.
            For fail-safe, events without "priority" property have the lowest priority.
            
            NOTICE: different from the queue priority, the priority property is smaller-higher, and is not limited to integers.
            This allows datetime to be used as an increasing priority
            '''
            def __init__(self, parent = None, maxlength = None, key = 'priority'):
                # a heap
                self.queue = []
                self.deque = deque()
                self.parent = parent
                self.maxlength = maxlength
                self.blocks = set()
                if self.maxlength is not None and self.maxlength <= 0:
                    self.maxlength = 1
                self.blocked = False
                self.isWaited = False
                self.key = key
            @classmethod
            def initHelper(cls, key = 'priority'):
                def initer(parent = None, maxlength = None):
                    return cls(parent, maxlength, key)
                return initer
            def append(self, value, force = False):
                if not force and not self.canAppend():
                    self.isWaited = True
                    return QueueCanWriteEvent.createMatcher(self)
                if self.parent is not None:
                    m = self.parent.notifyAppend(self, force)
                    if m is not None:
                        return m
                if hasattr(value, self.key):
                    heappush(self.queue, (getattr(value, self.key), value))
                    # a priority push may change the block status
                    if self.blocked and not self.queue[0][1] in self.blocks:
                        self.blocked = False
                        if self.parent is not None:
                            self.parent.notifyBlock(self, False)
                else:
                    self.deque.append(value)
                return None
            def canAppend(self):
                return self.maxlength is None or len(self.queue) + len(self.deque) < self.maxlength
            def canPop(self):
                return len(self.queue) + len(self.deque) > 0 and not self.blocked
            def pop(self):
                ret = self._pop()
                if self.parent is not None:
                    pr = self.parent.notifyPop(self)
                    ret[1].extend(pr[0])
                    ret[2].extend(pr[1])
                return ret
            def _pop(self):
                if self.blocked:
                    raise IndexError('pop from a blocked queue')
                if self.queue:
                    ret = heappop(self.queue)[1]
                else:
                    ret = self.deque.popleft()
                if self.queue:
                    blocked = self.queue[0][1] in self.blocks
                elif self.deque:
                    blocked = self.deque[0] in self.blocks
                else:
                    blocked = False
                if self.blocked != blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
                if self.isWaited and self.canAppend():
                    self.isWaited = False
                    return (ret, [QueueCanWriteEvent(self)], [])
                else:
                    return (ret, [], [])
            def _clear(self):
                if self.blocks:
                    self.unblockall()
                del self.queue[:]
                self.deque.clear()
                if self.isWaited and self.canAppend():
                    self.isWaited = False
                    return ([QueueCanWriteEvent(self)], [])
                else:
                    return ([], [])
            def clear(self):
                l = len(self)
                ret = self._clear()
                if self.parent is not None:
                    pr = self.parent.notifyPop(self, l)
                    ret[0].extend(pr[0])
                    ret[1].extend(pr[1])
                return ret
            def __len__(self):
                return len(self.queue) + len(self.deque)
            def block(self, value):
                self.blocks.add(value)
                if self.parent is not None:
                    self.parent.notifyAppend(self, True)
                if hasattr(value, self.key):
                    heappush(self.queue, (getattr(value, self.key), value))
                else:
                    self.deque.appendleft(value)
                if self.queue:
                    blocked = self.queue[0][1] in self.blocks
                elif self.deque:
                    blocked = self.deque[0] in self.blocks
                else:
                    blocked = False
                if self.blocked != blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
            def unblock(self, value):
                self.blocks.remove(value)
                if self.queue:
                    blocked = self.queue[0][1] in self.blocks
                elif self.deque:
                    blocked = self.deque[0] in self.blocks
                else:
                    blocked = False
                if self.blocked != blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
            def unblockall(self):
                self.blocks.clear()
                blocked = False
                if self.blocked != blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
        class MultiQueue(object):
            '''
            A multi-queue container, every queue in a multi-queue has the same priority, and is popped in turn.
            '''
            class CircleListNode(object):
                '''
                Circle link list
                '''
                def __init__(self, value):
                    self.prev = self
                    self.value = value
                    self.next = self
                def insertprev(self, node):
                    self.prev.next = node
                    node.prev = self.prev
                    node.next = self
                    self.prev = node
                    return self
                def remove(self):
                    if self.next is self:
                        return None
                    self.prev.next = self.next
                    self.next.prev = self.prev
                    ret = self.next
                    self.next = self
                    self.prev = self
                    return ret
            class CircleList(object):
                def __init__(self):
                    self.current = None
                def remove(self, node):
                    if self.current is node:
                        self.current = node.remove()
                    else:
                        node.remove()
                def insertprev(self, node):
                    if self.current is None:
                        self.current = node
                    else:
                        self.current.insertprev(node)
                def insertcurrent(self, node):
                    self.insertprev(node)
                    self.current = node
                def next(self):
                    ret = self.current
                    if self.current is not None:
                        self.current = self.current.next
                    return ret
                def clear(self):
                    self.current = None
            def __init__(self, parent = None, priority = 0):
                self.queues = CBQueue.MultiQueue.CircleList()
                self.queueDict = {}
                self.queueStat = {}
                self.statseq = deque()
                self.parent = parent
                self.priority = priority
                self.totalSize = 0
                self.blocked = True
            def canPop(self):
                return bool(self.queues.current)
            def _pop(self):
                if not self.canPop():
                    raise IndexError('pop from an empty or blocked queue')
                c = self.queues.next()
                ret = c.value._pop()
                self.queueStat[c.value] = self.queueStat.get(c.value, 0) + 1
                while len(self.statseq) >= 10 * len(self.queueDict) + 10:
                    o = self.statseq.popleft()
                    if o in self.queueStat:
                        self.queueStat[o] = self.queueStat[o] - 1
                        if self.queueStat[o] <= 0 and not o in self.queueDict:
                            del self.queueStat[o]
                self.statseq.append(c.value)
                if not c.value.canPop():
                    self.queues.remove(c)
                    self.queueDict[c.value] = None
                self.totalSize = self.totalSize - 1
                if not self.canPop():
                    if not self.blocked:
                        self.blocked = True
                        if self.parent is not None:
                            self.parent.notifyBlock(self, True)
                return ret            
            def addSubQueue(self, queue):
                self.totalSize = self.totalSize + len(queue)
                queue.parent = self
                if queue.canPop():
                    # Activate this queue
                    node = CBQueue.MultiQueue.CircleListNode(queue)
                    self.queues.insertprev(node)
                    self.queueDict[queue] = node
                    self.queueStat[queue] = 0
                else:
                    self.queueDict[queue] = None
                if self.canPop():
                    if self.blocked:
                        self.blocked = False
                        if self.parent is not None:
                            self.parent.notifyBlock(self, False)
            def removeSubQueue(self, queue):
                self.totalSize = self.totalSize - len(queue)
                if self.queueDict[queue] is not None:
                    self.queues.remove(self.queueDict[queue])
                del self.queueDict[queue]
                if queue in self.queueStat:
                    del self.queueStat[queue]
                if not self.canPop():
                    if not self.blocked:
                        self.blocked = True
                        if self.parent is not None:
                            self.parent.notifyBlock(self, True)
            def notifyAppend(self, queue, force):
                if self.parent is not None:
                    m = self.parent.notifyAppend(self, force)
                    if m is not None:
                        return m
                self.totalSize = self.totalSize + 1
                if not queue.blocked:
                    if self.queueDict[queue] is None:
                        # Activate this queue
                        node = CBQueue.MultiQueue.CircleListNode(queue)
                        qs = self.queueStat.setdefault(queue, 0)
                        if qs * len(self.queueStat) >= len(self.statseq):
                            self.queues.insertprev(node)
                        else:
                            self.queues.insertcurrent(node)
                        self.queueDict[queue] = node
                if self.canPop():
                    if self.blocked:
                        self.blocked = False
                        if self.parent is not None:
                            self.parent.notifyBlock(self, False)
                return None
            def __len__(self):
                return self.totalSize
            def notifyBlock(self, queue, blocked):
                if queue.canPop():
                    if self.queueDict[queue] is None:
                        # Activate this queue
                        node = CBQueue.MultiQueue.CircleListNode(queue)
                        qs = self.queueStat.setdefault(queue, 0)
                        if qs * len(self.queueStat) >= len(self.statseq):
                            self.queues.insertprev(node)
                        else:
                            self.queues.insertcurrent(node)
                        self.queueDict[queue] = node
                else:
                    if self.queueDict[queue] is not None:
                        self.queues.remove(self.queueDict[queue])
                        self.queueDict[queue] = None
                selfblocked = not self.canPop()
                if selfblocked != self.blocked:
                    self.blocked = selfblocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, selfblocked)
            def notifyPop(self, queue, length = 1):
                self.totalSize = self.totalSize - length
                if not queue.canPop():
                    if self.queueDict[queue] is not None:
                        self.queues.remove(self.queuDict[queue])
                        self.queueDict[queue] = None
                ret = ([], [])
                if self.parent is not None:
                    ret = self.parent.notifyPop(self, length)
                blocked = not self.canPop()
                if blocked != self.blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
                return ret
                
            def unblockall(self):
                for q in self.queueDict.keys():
                    q.unblockall()
            def _clear(self):
                ret = ([],[])
                for q in self.queueDict.keys():
                    pr = q._clear()
                    ret[0].extend(pr[0])
                    ret[1].extend(pr[1])
                self.totalSize = 0
                self.blockedSize = 0
                self.queues.clear()
                if not self.blocked:
                    self.blocked = True
                    if self.parent is not None:
                        self.parent.notifyBlock(self, True)
                return ret
        class AutoClassQueue(object):
            '''
            A queue classify events into virtual sub-queues by key
            '''
            nokey = object()
            def __init__(self, parent = None, maxlength = None, key = 'owner', preserveForNew = 1, maxstat = None, subqueuelimit = None):
                self.queues = CBQueue.MultiQueue.CircleList()
                self.queueDict = {}
                self.queueStat = {}
                self.statseq = deque()
                self.maxlength = maxlength
                self.blocked = False
                if self.maxlength is not None and self.maxlength <= 0:
                    self.maxlength = 1
                if maxstat is None:
                    if self.maxlength is None:
                        self.maxstat = 10240
                    else:
                        self.maxstat = maxlength * 10
                else:
                    self.maxstat = maxstat
                if self.maxstat >= 10240:
                    self.maxstat = 10240
                self.waited = set()
                self.key = key
                self.preserve = preserveForNew
                self.totalSize = 0
                self.blockKeys = set()
                self.subqueuelimit = subqueuelimit
            @classmethod
            def initHelper(cls, key = 'owner', preserveForNew = 1, maxstat = None, subqueuelimit = None):
                def initer(parent = None, maxlength = None):
                    return cls(parent, maxlength, key, preserveForNew, maxstat, subqueuelimit)
                return initer
            def append(self, value, force = False):
                key = getattr(value, self.key, self.nokey)
                # We use hash instead of reference or weakref, this may cause problem, but better thank leak.
                kid = hash(key)
                if not force:
                    w = self._tryAppend(key)
                    if w is not None:
                        return w
                if self.parent is not None:
                    m = self.parent.notifyAppend(self, force)
                    if m is not None:
                        return m
                if key in self.queueDict:
                    self.queueDict[key].value[1].append(value)
                else:
                    node = CBQueue.MultiQueue.CircleListNode((key,deque()))
                    node.value[1].append(value)
                    qs = self.queueStat.setdefault(kid, 0)
                    if qs * len(self.queueStat) >= len(self.statseq):
                        self.queues.insertprev(node)
                    else:
                        self.queues.insertcurrent(node)
                    self.queueDict[key] = node
                self.totalSize += 1
                blocked = not self.canPop() and self.totalSize > 0
                if blocked != self.blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
                return None
            def _tryAppend(self, key):
                if self.maxlength is None:
                    if self.subqueuelimit is None or not key in self.queueDict:
                        return None
                    elif len(self.queueDict[key].value[1]) >= self.subqueuelimit:
                        self.waited.add((False, False, key))
                        return AutoClassQueueCanWriteEvent.createMatcher(self, _ismatch = lambda x: x.key == key or x.key is self.nokey)
                    else:
                        return None
                if key in self.queueDict:
                    if self.subqueuelimit is not None and len(self.queueDict[key].value[1]) >= self.subqueuelimit:
                        self.waited.add((False, False, key))
                        return AutoClassQueueCanWriteEvent.createMatcher(self, _ismatch = lambda x: x.key == key or x.key is self.nokey)
                    elif self.totalSize < self.maxlength - self.preserve - len(self.queueStat) + len(self.queueDict):
                        return None
                    else:
                        if len(self.queueDict[key].value[1]) <= 1:
                            self.waited.add((False, True, key))
                            return AutoClassQueueCanWriteEvent.createMatcher(self, False, _ismatch = lambda x: not (x.firstonly and x.key != key))
                        else:
                            self.waited.add((False, False))
                            return AutoClassQueueCanWriteEvent.createMatcher(self, False, False)
                elif hash(key) in self.queueStat:
                    if self.totalSize < self.maxlength - self.preserve:
                        return None
                    else:
                        self.waited.add((False, True))
                        return AutoClassQueueCanWriteEvent.createMatcher(self, False)
                else:
                    if self.totalSize < self.maxlength:
                        return None
                    else:
                        self.waited.add((True, True))
                        return AutoClassQueueCanWriteEvent.createMatcher(self)
            def canAppend(self):
                return self.maxlength is None or self.totalSize < self.maxlength
            def canPop(self):
                return self.queues.current is not None
            def pop(self):
                ret = self._pop()
                if self.parent is not None:
                    pr = self.parent.notifyPop(self)
                    ret[1].extend(pr[0])
                    ret[2].extend(pr[1])
                return ret
            def _pop(self):
                if not self.canPop():
                    raise IndexError('pop from a blocked or empty queue')
                c = self.queues.next()
                key = c.value[0]
                kid = hash(key)
                ret = c.value[1].popleft()
                self.totalSize -= 1
                self.queueStat[kid] = self.queueStat.get(kid, 0) + 1
                while len(self.statseq) >= min(self.maxstat, 10 * len(self.queueStat) + 10):
                    k1 = self.statseq.popleft()
                    self.queueStat[k1] = self.queueStat[k1] - 1
                    if self.queueStat[k1] <= 0:
                        del self.queueStat[k1]
                self.statseq.append(kid)
                if not c.value[1]:
                    del self.queueDict[c.value[0]]
                    self.queues.remove(c)
                blocked = not self.canPop() and self.totalSize > 0
                if blocked != self.blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
                if self.waited:
                    if key not in self.queueDict:
                        subsize = 0
                    else:
                        subsize = len(self.queueDict[key].value[1])
                    if self.maxlength is None:
                        if self.subqueuelimit is not None and subsize < self.subqueuelimit and (False, False, key) in self.waited:
                            return (ret, [AutoClassQueueCanWriteEvent(self, False, False, key=key)], [])
                    elif self.totalSize < self.maxlength - self.preserve - len(self.queueStat) + len(self.queueDict):
                        self.waited = set(w for w in self.waited if len(w) == 3 and w[1] == False and w[2] != key)
                        return (ret, [AutoClassQueueCanWriteEvent(self, False, False, key=key)], [])
                    elif self.totalSize < self.maxlength - self.preserve:
                        if (False, True) in self.waited or (False, True, key) in self.waited or (True, True) in self.waited or \
                                (False, False, key) in self.waited:
                            self.waited.discard((False, True))
                            self.waited.discard((False, True, key))
                            self.waited.discard((True, True))
                            self.waited.discard((False, False, key))
                            return (ret, [AutoClassQueueCanWriteEvent(self, False, True, key=key)], [])
                    elif self.totalSize < self.maxlength:
                        if (True, True) in self.waited or (False, False, key) in self.waited or (False, True, key) in self.waited:
                            self.waited.discard((True, True))
                            self.waited.discard((False, False, key))
                            if (False, True, key) in self.waited:
                                self.waited.discard((False, True, key))
                                return (ret, [AutoClassQueueCanWriteEvent(self, False, True, key=key)], [])
                            else:
                                return (ret, [AutoClassQueueCanWriteEvent(self, True, True, key=key)], [])
                    elif self.subqueuelimit is not None and subsize < self.subqueuelimit and (False, False, key) in self.waited:
                        # If we don't wake up the sub-queue waiter now, it may wait forever.
                        # The sub-queue waiter won't be able to send events in, but they will get a new matcher
                        # Some waiters might wake up mistakenly, they will wait again when they try to append the event. 
                        self.waited.discard((True, True))
                        self.waited.discard((False, False, key))
                        return (ret, [AutoClassQueueCanWriteEvent(self, True, True, key=key)], [])
                return (ret, [], [])
            def clear(self):
                l = len(self)
                ret = self._clear()
                if self.parent is not None:
                    pr = self.parent.notifyPop(self, l)
                    ret[0].extend(pr[0])
                    ret[1].extend(pr[1])
                return ret
            def _clear(self):
                self.queues.clear()
                self.blockKeys.clear()
                self.queueDict.clear()
                self.totalSize = 0
                blocked = not self.canPop() and self.totalSize > 0
                if blocked != self.blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
                if self.waited:
                    self.waited.clear()
                    return ([AutoClassQueueCanWriteEvent(self, False, False, key=self.nokey)], [])
                else:
                    return ([], [])
            def __len__(self):
                return self.totalSize
            def block(self, value):
                if self.parent is not None:
                    self.parent.notifyAppend(self, True)
                key = getattr(value, self.key, self.nokey)
                if key in self.queueDict:
                    self.queueDict[key].value[1].appendleft(value)
                    self.queues.remove(self.queueDict[key])
                else:
                    node = CBQueue.MultiQueue.CircleListNode((key,deque()))
                    node.value[1].append(value)
                    self.queueDict[key] = node
                self.blockKeys.add(key)
                self.totalSize += 1
                blocked = not self.canPop() and self.totalSize > 0
                if blocked != self.blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
            def unblock(self, value):
                key = getattr(value, self.key, self.nokey)
                if key in self.blockKeys:
                    self._unblock(key)
                    blocked = not self.canPop() and self.totalSize > 0
                    if blocked != self.blocked:
                        self.blocked = blocked
                        if self.parent is not None:
                            self.parent.notifyBlock(self, blocked)
            def _unblock(self, key):
                self.blockKeys.remove(key)
                node = self.queueDict[key]
                qs = self.queueStat.setdefault(hash(key), 0)
                if qs * len(self.queueStat) >= len(self.statseq):
                    self.queues.insertprev(node)
                else:
                    self.queues.insertcurrent(node)            
            def unblockall(self):
                for k in list(self.blockKeys):
                    self._unblock(k)
                blocked = not self.canPop() and self.totalSize > 0
                if blocked != self.blocked:
                    self.blocked = blocked
                    if self.parent is not None:
                        self.parent.notifyBlock(self, blocked)
        def __init__(self, tree = None, parent = None, maxdefault = None, maxtotal = None, defaultQueueClass = FifoQueue, defaultQueuePriority = 0):
            '''
            Constructor
            '''
            self.queues = {}
            self.queueindex = {}
            self.prioritySet = []
            if tree is None:
                self.tree = MatchTree()
            else:
                self.tree = tree
            self.parent = parent
            defaultPriority = CBQueue.MultiQueue(self, defaultQueuePriority)
            defaultQueue = defaultQueueClass(defaultPriority, maxdefault)
            defaultPriority.addSubQueue(defaultQueue)
            self.queues[defaultQueuePriority] = defaultPriority
            self.tree.insert(None, defaultQueue)
            self.defaultQueue = defaultQueue
            self.totalSize = 0
            self.maxtotal = maxtotal
            self.blocked = True
            self.blockEvents = {}
            self.isWaited = False
            self.isWaitEmpty = False
            self.outputStat = 0
        def _removeFromTree(self):
            for v in self.queueindex.values():
                if len(v) == 3:
                    v[1]._removeFromTree()
            self.tree.remove(None, self.defaultQueue)
            self.tree = None
        def canAppend(self):
            '''
            Whether the queue is full or not. Only check the total limit. Sub-queue may still be full (even default).
            
            :returns: False if the queue is full, True if not.
                      If there are sub-queues, append() may still fail if the sub-queue is full. 
            '''
            return self.maxtotal is None or self.totalSize < self.maxtotal
        def append(self, event, force = False):
            '''
            Append an event to queue. The events are classified and appended to sub-queues
            
            :param event: input event
            
            :param force: if True, the event is appended even if the queue is full
            
            :returns: None if appended successfully, or a matcher to match a QueueCanWriteEvent otherwise
            '''
            if self.tree is None:
                if self.parent is None:
                    raise IndexError('The queue is removed')
                else:
                    return self.parent.parent.append(event, force)
            q = self.tree.matchfirst(event)
            return q.append(event, force)
        def waitForEmpty(self):
            '''
            Make this queue generate a QueueIsEmptyEvent when it is empty
            
            :returns: matcher for QueueIsEmptyEvent, or None if the queue is already empty
            '''
            if not self:
                return None
            self.isWaitEmpty = True
            return QueueIsEmptyEvent.createMatcher(self)
        def block(self, event, emptyEvents = ()):
            '''
            Return a recently popped event to queue, and block all later events until unblock.
            
            Only the sub-queue directly containing the event is blocked, so events in other queues may still be processed.
            It is illegal to call block and unblock in different queues with a same event.
            
            :param event: the returned event. When the queue is unblocked later, this event will be popped again.
            
            :param emptyEvents: reactivate the QueueIsEmptyEvents
            '''
            q = self.tree.matchfirst(event)
            q.block(event)
            self.blockEvents[event] = q
            for ee in emptyEvents:
                ee.queue.waitForEmpty()
        def unblock(self, event):
            '''
            Remove a block 
            '''
            if event not in self.blockEvents:
                return
            self.blockEvents[event].unblock(event)
            del self.blockEvents[event]
        def unblockqueue(self, queue):
            '''
            Remove blocked events from the queue and all subqueues. Usually used after queue clear/unblockall to prevent leak.
            
            :returns: the cleared events
            '''
            subqueues = set()
            def allSubqueues(q):
                subqueues.add(q)
                subqueues.add(q.defaultQueue)
                for v in q.queueindex.values():
                    if len(v) == 3:
                        allSubqueues(v[1])
            allSubqueues(queue)
            events = [k for k,v in self.blockEvents.items() if v in subqueues]
            for e in events:
                del self.blockEvents[e]
            return events
        def unblockall(self):
            '''
            Remove all blocks from the queue and all sub-queues
            '''
            for q in self.queues.values():
                q.unblockall()
            self.blockEvents.clear()
        def notifyAppend(self, queue, force):
            '''
            Internal notify for sub-queues
            
            :returns: If the append is blocked by parent, an EventMatcher is returned, None else.
            '''
            if not force and not self.canAppend():
                self.isWaited = True
                return QueueCanWriteEvent.createMatcher(self)
            if self.parent is not None:
                m = self.parent.notifyAppend(self, force)
                if m is not None:
                    return m
            self.totalSize = self.totalSize + 1
            return None
        def notifyBlock(self, queue, blocked):
            '''
            Internal notify for sub-queues been blocked
            '''
            if blocked:
                if self.prioritySet[-1] == queue.priority:
                    self.prioritySet.pop()
                else:
                    pindex = bisect_left(self.prioritySet, queue.priority)
                    if pindex < len(self.prioritySet) and self.prioritySet[pindex] == queue.priority:
                        del self.prioritySet[pindex]
            else:
                if queue.canPop():
                    pindex = bisect_left(self.prioritySet, queue.priority)
                    if pindex >= len(self.prioritySet) or self.prioritySet[pindex] != queue.priority:
                        self.prioritySet.insert(pindex, queue.priority)
            newblocked =  not self.canPop()
            if newblocked != self.blocked:
                self.blocked = newblocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, newblocked)
        def notifyPop(self, queue, length = 1):
            '''
            Internal notify for sub-queues been poped
            
            :returns: List of any events generated by this pop
            '''
            self.totalSize = self.totalSize - length
            ret1 = []
            ret2 = []
            if self.isWaited and self.canAppend():
                self.isWaited = False
                ret1.append(QueueCanWriteEvent(self))
            if self.isWaitEmpty and not self:
                self.isWaitEmpty = False
                ret2.append(QueueIsEmptyEvent(self))
            if self.parent is not None:
                pr = self.parent.notifyPop(self, length)
                ret1 += pr[0]
                ret2 += pr[1]
            newblocked =  not self.canPop()
            if newblocked != self.blocked:
                self.blocked = newblocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, newblocked)
            return (ret1, ret2)
        def canPop(self):
            '''
            Whether the queue is empty/blocked or not
            
            :returns: False if the queue is empty or blocked, or True otherwise
            '''
            return bool(self.prioritySet)
        def pop(self):
            '''
            Pop an event from the queue. The event in the queue with higher priority is popped before ones in lower priority.
            If there are multiple queues with the same priority, events are taken in turn from each queue.
            May return some queueEvents indicating that some of the queues can be written into.
            
            :returns: `(obj, (queueEvents,...), (queueEmptyEvents,...))` where obj is the popped event, queueEvents are QueueCanWriteEvents generated by this pop
                      and queueEmptyEvents are QueueIsEmptyEvents generated by this pop
            '''
            ret = self._pop()
            if self.parent is not None:
                pr = self.parent.notifyPop(self)
                ret[1].extend(pr[0])
                ret[2].extend(pr[1])
            return ret
        def _pop(self):
            '''
            Actual pop
            '''
            if not self.canPop():
                raise IndexError('pop from an empty or blocked queue')
            priority = self.prioritySet[-1]
            ret = self.queues[priority]._pop()
            self.outputStat = self.outputStat + 1
            self.totalSize = self.totalSize - 1
            if self.isWaited and self.canAppend():
                self.isWaited = False
                ret[1].append(QueueCanWriteEvent(self))
            if self.isWaitEmpty and not self:
                self.isWaitEmpty = False
                ret[2].append(QueueIsEmptyEvent(self))
            return ret
        def clear(self):
            '''
            Clear all the events in this queue, including any sub-queues.
            
            :returns: ((queueEvents,...), (queueEmptyEvents,...)) where queueEvents are QueueCanWriteEvents generated by clearing.
            '''
            l = len(self)
            ret = self._clear()
            if self.parent is not None:
                pr = self.parent.notifyPop(self, l)
                ret[0].extend(pr[0])
                ret[1].extend(pr[1])
            return ret
        def _clear(self):
            '''
            Actual clear
            '''
            ret = ([],[])
            for q in self.queues.values():
                pr = q._clear()
                ret[0].extend(pr[0])
                ret[1].extend(pr[1])
            self.totalSize = 0
            del self.prioritySet[:]
            if self.isWaited and self.canAppend():
                self.isWaited = False
                ret[0].append(QueueCanWriteEvent(self))
            if self.isWaitEmpty and not self:
                self.isWaitEmpty = False
                ret[1].append(QueueIsEmptyEvent(self))
            self.blockEvents.clear()
            return ret
        def __contains__(self, name):
            return name in self.queueindex
        def __getitem__(self, name):
            '''
            Get a sub-queue through q['sub-queue-name']
            '''
            return self.queueindex[name][1]
        def getPriority(self, queue):
            '''
            get priority of a sub-queue
            '''
            return self.queueindex[queue][0]
        def setPriority(self, queue, priority):
            '''
            Set priority of a sub-queue
            '''
            q = self.queueindex[queue]
            self.queues[q[0]].removeSubQueue(q[1])
            newPriority = self.queues.setdefault(priority, CBQueue.MultiQueue(self, priority))
            q[0] = priority
            newPriority.addSubQueue(q[1])
            
        def addSubQueue(self, priority, matcher, name = None, maxdefault = None, maxtotal = None, defaultQueueClass = FifoQueue):
            '''
            add a sub queue to current queue, with a priority and a matcher
            
            :param priority: priority of this queue. Larger is higher, 0 is lowest.
            
            :param matcher: an event matcher to catch events. Every event match the criteria will be stored in this queue.
            
            :param name: a unique name to identify the sub-queue. If none, the queue is anonymous. It can be any hashable value.
            
            :param maxdefault: max length for default queue.
            
            :param maxtotal: max length for sub-queue total, including sub-queues of sub-queue
            '''
            if name is not None and name in self.queueindex:
                raise IndexError("Duplicated sub-queue name '" + str(name) + "'")
            subtree = self.tree.subtree(matcher, True)
            newPriority = self.queues.setdefault(priority, CBQueue.MultiQueue(self, priority))
            newQueue = CBQueue(subtree, newPriority, maxdefault, maxtotal, defaultQueueClass)
            newPriority.addSubQueue(newQueue)
            qi = [priority, newQueue, name]
            if name is not None:
                self.queueindex[name] = qi
            self.queueindex[newQueue] = qi
            return newQueue
        def removeSubQueue(self, queue):
            '''
            remove a sub queue from current queue.
            
            This unblock the sub-queue, retrieve all events from the queue and put them back to the parent.
            
            Call clear on the sub-queue first if the events are not needed any more.
            
            :param queue: the name or queue object to remove
            
            :returns: ((queueevents,...), (queueEmptyEvents,...)) Possible queue events from removing sub-queues
            '''
            q = self.queueindex[queue]
            q[1].unblockall()
            q[1]._removeFromTree()
            ret = ([],[])
            while q[1].canPop():
                r = q[1].pop()
                self.append(r[0], True)
                ret[0].extend(r[1])
                ret[1].extend(r[2])
            self.queues[q[0]].removeSubQueue(q[1])
            # Remove from index
            if q[2] is not None:
                del self.queueindex[q[2]]
            del self.queueindex[q[1]]
            newblocked =  not self.canPop()
            if newblocked != self.blocked:
                self.blocked = newblocked
                if self.parent is not None:
                    self.parent.notifyBlock(self, newblocked)
            return ret
        def __len__(self):
            return self.totalSize
            



'''
Created on 2015/8/14

:author: hubo
'''

from vlcp.event import Event, withIndices
import os

@withIndices('stream', 'type')
class StreamDataEvent(Event):
    canignore = False
    STREAM_DATA = 'data'
    STREAM_EOF = 'eof'
    STREAM_ERROR = 'error'
    def canignorenow(self):
        return self.stream.closed

class BaseStream(object):
    "Streaming base"
    def __init__(self, isunicode = False, encoders = []):
        '''
        Constructor
        
        :param isunicode: if True, the data used in this stream outputs unicode string. Otherwise it outputs bytes.
        
        :param encoders: a list of functions `enc(data, is_eof)` which encodes input data.
        '''
        self.data = b''
        self.pos = 0
        self.dataeof = False
        self.dataerror = False
        self.eof = False
        self.errored = False
        self.closed = False
        self.writeclosed = False
        self.allowwrite = False
        self.isunicode = isunicode
        self.encoders = list(encoders)
    def getEncoderList(self):
        """
        Return the encoder list
        """
        return self.encoders
    def _parsedata(self, data, dataeof, dataerror):
        self.pos = 0
        self.dataeof = dataeof
        self.dataerror = dataerror
        if len(data) > 0 or self.dataeof:
            for enc in self.encoders:
                try:
                    data = enc(data, self.dataeof)
                except:
                    self.errored = True
                    self.closed = True
                    raise
        self.data = data
    def read(self, container, size = None):
        """
        Coroutine method to read from the stream and return the data from `container.data`. Raises EOFError
        when the stream end has been reached; raises IOError if there are other errors.
        
        :param container: A routine container
        
        :param size: maximum read size, or unlimited if is None
        """
        ret = []
        retsize = 0
        if self.eof:
            raise EOFError
        if self.errored:
            raise IOError('Stream is broken before EOF')
        while size is None or retsize < size:
            if self.pos >= len(self.data):
                for m in self.prepareRead(container):
                    yield m
            if size is None or size - retsize >= len(self.data) - self.pos:
                t = self.data[self.pos:]
                ret.append(t)
                retsize += len(t)
                self.pos = len(self.data)
                if self.dataeof:
                    self.eof = True
                    break
                if self.dataerror:
                    self.errored = True
                    break
            else:
                ret.append(self.data[self.pos:self.pos + (size - retsize)])
                self.pos += (size - retsize)
                retsize = size
                break
        if self.isunicode:
            container.data = u''.join(ret)
        else:
            container.data = b''.join(ret)
        if self.errored:
            raise IOError('Stream is broken before EOF')
    def readonce(self, size = None):
        """
        Read from current buffer. If current buffer is empty, returns an empty string. You can use `prepareRead`
        to read the next chunk of data.
        
        This is not a coroutine method.
        """
        if self.eof:
            raise EOFError
        if self.errored:
            raise IOError('Stream is broken before EOF')
        if size is not None and size < len(self.data) - self.pos:
            ret = self.data[self.pos: self.pos + size]
            self.pos += size
            return ret
        else:
            ret = self.data[self.pos:]
            self.pos = len(self.data)
            if self.dataeof:
                self.eof = True
            return ret
    def prepareRead(self, container):
        """
        A coroutine method to read the next chunk of data.
        """
        if False:
            yield
        raise NotImplementedError
    def readline(self, container, size = None):
        """
        Coroutine method which reads the next line or until EOF or size exceeds
        """
        ret = []
        retsize = 0
        if self.eof:
            raise EOFError
        if self.errored:
            raise IOError('Stream is broken before EOF')
        while size is None or retsize < size:
            if self.pos >= len(self.data):
                for m in self.prepareRead(container):
                    yield m
            if size is None or size - retsize >= len(self.data) - self.pos:
                t = self.data[self.pos:]
                if self.isunicode:
                    p = t.find(u'\n')
                else:
                    p = t.find(b'\n')
                if p >= 0:
                    t = t[0: p + 1]
                    ret.append(t)
                    retsize += len(t)
                    self.pos += len(t)
                    break
                else:
                    ret.append(t)
                    retsize += len(t)
                    self.pos += len(t)
                    if self.dataeof:
                        self.eof = True
                        break
                    if self.dataerror:
                        self.errored = True
                        break
            else:
                t = self.data[self.pos:self.pos + (size - retsize)]
                if self.isunicode:
                    p = t.find(u'\n')
                else:
                    p = t.find(b'\n')
                if p >= 0:
                    t = t[0: p + 1]
                ret.append(t)
                self.pos += len(t)
                retsize += len(t)
                break
        if self.isunicode:
            container.data = u''.join(ret)
        else:
            container.data = b''.join(ret)
        if self.errored:
            raise IOError('Stream is broken before EOF')
    def copyTo(self, dest, container, buffering = True):
        """
        Coroutine method to copy content from this stream to another stream.
        """
        if self.eof:
            for m in dest.write(u'' if self.isunicode else b'', True):
                yield m
        elif self.errored:
            for m in dest.error(container):
                yield m
        else:
            try:
                while not self.eof:
                    for m in self.prepareRead(container):
                        yield m
                    data = self.readonce()
                    try:
                        for m in dest.write(data, container, self.eof, buffering = buffering):
                            yield m
                    except IOError:
                        break
            except:
                try:
                    for m in dest.error(container):
                        yield m
                except IOError:
                    pass
                raise
            finally:
                self.close(container.scheduler)
    def write(self, data, container, eof = False, ignoreexception = False, buffering = True, split = True):
        """
        Coroutine method to write data to this stream.
        
        :param data: data to write
        
        :param container: the routine container
        
        :param eof: if True, this is the last chunk of this stream. The other end will receive an EOF after reading
                    this chunk.
                    
        :param ignoreexception: even if the stream is closed on the other side, do not raise exception.
        
        :param buffering: enable buffering. The written data may not be sent if buffering = True; if buffering = False,
                          immediately send any data in the buffer together with this chunk.
        
        :param split: enable splitting. If this chunk is too large, the stream is allowed to split it into
                      smaller chunks for better balancing.
        """
        if not ignoreexception:
            raise IOError('Stream is closed')
        if False:
            yield
    def error(self, container, ignoreexception = False):
        """
        Raises error on this stream, so that the receiving end gets an IOError exception.
        """
        if not ignoreexception:
            raise IOError('Stream is closed')
        if False:
            yield
    def close(self, scheduler = None, allowwrite = False):
        """
        Read side close. To close at the write side, use `eof=True` with `write`.
        
        :param scheduler: the scheduler
        
        :param allowwrite: do not raise exception on the write side
        """
        self.closed = True

class Stream(BaseStream):
    '''
    Streaming data with events
    '''
    def __init__(self, isunicode = False, encoders = [], writebufferlimit = 4096, splitsize = 1048576):
        '''
        Constructor
        
        :param isunicode: True if this stream outputs unicode; False if this stream outputs bytes
        
        :param encoders: a list of functions `enc(data, is_eof)` which encodes input data.
        
        :param writebufferlimit: if `buffering=True` on `write`, do not send data until there is more data than this limit
                                 
        :param splitsize: if `split=True` on `write`, split chunks larger than this to chunks with this size
        '''
        BaseStream.__init__(self, isunicode, encoders)
        self.dm = StreamDataEvent.createMatcher(self, None)
        self.writebuffer = []
        self.writebuffersize = 0
        self.writebufferlimit = writebufferlimit
        self.splitsize = splitsize
    def prepareRead(self, container):
        if not self.eof and not self.errored and self.pos >= len(self.data):
            yield (self.dm,)
            container.event.canignore = True
            self._parsedata(container.event.data, container.event.type == StreamDataEvent.STREAM_EOF,
                            container.event.type == StreamDataEvent.STREAM_ERROR)
    def write(self, data, container, eof = False, ignoreexception = False, buffering = True, split = True):
        if self.closed or self.writeclosed:
            if not ignoreexception and not self.allowwrite:
                raise IOError('Stream is closed')
        else:
            if eof:
                buffering = False
            if buffering and self.writebufferlimit is not None and self.writebuffersize + len(data) > self.writebufferlimit:
                buffering = False
            if buffering:
                self.writebuffer.append(data)
                self.writebuffersize += len(data)
            else:
                data = data[0:0].join(self.writebuffer) + data
                if not data:
                    split = False
                del self.writebuffer[:]
                self.writebuffersize = 0
                if split:
                    for i in range(0, len(data), self.splitsize):
                        for m in container.waitForSend(StreamDataEvent(self,
                                                                       StreamDataEvent.STREAM_EOF if eof and i + self.splitsize >= len(data) else StreamDataEvent.STREAM_DATA,
                                                                       data = data[i : i + self.splitsize])):
                            yield m                    
                else:
                    for m in container.waitForSend(StreamDataEvent(self,
                                                                   StreamDataEvent.STREAM_EOF if eof else StreamDataEvent.STREAM_DATA,
                                                                   data = data)):
                        yield m
                if eof:
                    self.writeclosed = True
    def error(self, container, ignoreexception = False):
        if self.closed or self.writeclosed:
            if not ignoreexception and not self.allowwrite:
                raise IOError('Stream is closed')
        else:
            for m in container.waitForSend(StreamDataEvent(self, StreamDataEvent.STREAM_ERROR, data = b'')):
                yield m
            self.writeclosed = True
    def close(self, scheduler, allowwrite = False):
        self.closed = True
        self.allowwrite = allowwrite
        scheduler.ignore(self.dm)

class MemoryStream(BaseStream):
    '''
    A stream with readonly data
    '''
    def __init__(self, data, encoders= [], isunicode = None):
        '''
        Constructor
        
        :param data: all input data
        
        :param encoders: encoder list
        
        :param isunicode: Whether this stream outputs unicode. Default to be the same to data. Notice that
                          the encoders may change bytes data to unicode or vice-versa
        '''
        BaseStream.__init__(self, not isinstance(data, bytes) if isunicode is None else isunicode, encoders)
        self.preloaddata = data
    def __len__(self):
        return len(self.preloaddata)
    def prepareRead(self, container):
        if not self.eof and not self.errored and self.pos >= len(self.data):
            self._parsedata(self.preloaddata, True, False)
        if False:
            yield

class FileStream(BaseStream):
    "A stream from a file-like object. The file-like object must be in blocking mode"
    def __init__(self, fobj, encoders = [], isunicode = None, size = None, readlimit = 65536):
        "Constructor"
        if isunicode is None:
            mode = getattr(fobj, 'mode', 'rb')
            if 'b' in mode:
                isunicode = False
            elif str is bytes:
                isunicode = False
            else:
                isunicode = True
        BaseStream.__init__(self, isunicode, encoders)
        self.fobj = fobj
        self.size = None
        try:
            self.size = os.fstat(fobj.fileno()).st_size
            self.size -= fobj.tell()
            if self.size > size:
                self.size = size
        except:
            if size:
                self.size = size
        self.readlimit = readlimit
        self.totalbuffered = 0
    def __len__(self):
        if self.size is None:
            raise TypeError('Cannot determine file size')
        else:
            return self.size
    def prepareRead(self, container):
        if not self.eof and not self.errored and self.pos >= len(self.data):
            try:
                readlimit = self.readlimit
                if self.size is not None and self.size - self.totalbuffered < readlimit:
                    readlimit = self.size - self.totalbuffered 
                data = self.fobj.read(readlimit)
                self.totalbuffered += len(data)
                eof = not data
                if self.size is not None and self.totalbuffered >= self.size:
                    eof = True
            except:
                self._parsedata(b'', False, True)
            else:
                self._parsedata(data, eof, False)
        if False:
            yield
    def close(self, scheduler=None, allowwrite=False):
        self.fobj.close()
        BaseStream.close(self, scheduler=scheduler, allowwrite=allowwrite)

class FileWriter(object):
    "Write to file"
    def __init__(self, fobj):
        self.fobj = fobj
    def write(self, data, container, eof = False, ignoreexception = False, buffering = True, split = True):
        try:
            if data:
                self.fobj.write(data)
            if eof:
                self.fobj.close()
        except:
            if not ignoreexception:
                raise
        if False:
            yield
    def error(self, container, ignoreexception = False):
        try:
            self.fobj.close()
        except:
            if not ignoreexception:
                raise
        if False:
            yield



import logging

class ContextAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        if 'context' in self.extra:
            m, ka = logging.LoggerAdapter.process(self, msg, kwargs)
            return ('(%s) %s' % (self.extra['context'], m), ka)
        else:
            return logging.LoggerAdapter.process(self, msg, kwargs)
